# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import WarpPointsFromStd


def test_WarpPointsFromStd_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        coord_mm=dict(
            argstr="-mm",
            xor=["coord_vox"],
        ),
        coord_vox=dict(
            argstr="-vox",
            xor=["coord_mm"],
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        img_file=dict(
            argstr="-img %s",
            extensions=None,
            mandatory=True,
        ),
        in_coords=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        std_file=dict(
            argstr="-std %s",
            extensions=None,
            mandatory=True,
        ),
        warp_file=dict(
            argstr="-warp %s",
            extensions=None,
            xor=["xfm_file"],
        ),
        xfm_file=dict(
            argstr="-xfm %s",
            extensions=None,
            xor=["warp_file"],
        ),
    )
    inputs = WarpPointsFromStd.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_WarpPointsFromStd_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = WarpPointsFromStd.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..tracking import Tracks2Prob


def test_Tracks2Prob_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        colour=dict(
            argstr="-colour",
            position=3,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fraction=dict(
            argstr="-fraction",
            position=3,
        ),
        in_file=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=-2,
        ),
        out_filename=dict(
            argstr="%s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        output_datatype=dict(
            argstr="-datatype %s",
            position=2,
        ),
        resample=dict(
            argstr="-resample %d",
            position=3,
            units="mm",
        ),
        template_file=dict(
            argstr="-template %s",
            extensions=None,
            position=1,
        ),
        voxel_dims=dict(
            argstr="-vox %s",
            position=2,
            sep=",",
        ),
    )
    inputs = Tracks2Prob.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Tracks2Prob_outputs():
    output_map = dict(
        tract_image=dict(
            extensions=None,
        ),
    )
    outputs = Tracks2Prob.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



from ..preprocess import BBRegister, BBRegisterInputSpec6


def test_BBRegister_inputs():
    input_map_5_3 = dict(
        args=dict(argstr="%s",),
        contrast_type=dict(argstr="--%s", mandatory=True,),
        dof=dict(argstr="--%d",),
        environ=dict(nohash=True, usedefault=True,),
        epi_mask=dict(argstr="--epi-mask",),
        fsldof=dict(argstr="--fsl-dof %d",),
        init=dict(argstr="--init-%s", mandatory=True, xor=["init_reg_file"],),
        init_cost_file=dict(argstr="--initcost %s",),
        init_reg_file=dict(argstr="--init-reg %s", mandatory=True, xor=["init"],),
        intermediate_file=dict(argstr="--int %s",),
        out_fsl_file=dict(argstr="--fslmat %s",),
        out_lta_file=dict(argstr="--lta %s", min_ver="5.2.0",),
        out_reg_file=dict(argstr="--reg %s", genfile=True,),
        reg_frame=dict(argstr="--frame %d", xor=["reg_middle_frame"],),
        reg_middle_frame=dict(argstr="--mid-frame", xor=["reg_frame"],),
        registered_file=dict(argstr="--o %s",),
        source_file=dict(argstr="--mov %s", copyfile=False, mandatory=True,),
        spm_nifti=dict(argstr="--spm-nii",),
        subject_id=dict(argstr="--s %s", mandatory=True,),
        subjects_dir=dict(),
    )
    input_map_6_0 = dict(
        args=dict(argstr="%s",),
        contrast_type=dict(argstr="--%s", mandatory=True,),
        dof=dict(argstr="--%d",),
        environ=dict(nohash=True, usedefault=True,),
        epi_mask=dict(argstr="--epi-mask",),
        fsldof=dict(argstr="--fsl-dof %d",),
        init=dict(argstr="--init-%s", xor=["init_reg_file"],),
        init_reg_file=dict(argstr="--init-reg %s", xor=["init"],),
        init_cost_file=dict(argstr="--initcost %s",),
        intermediate_file=dict(argstr="--int %s",),
        out_fsl_file=dict(argstr="--fslmat %s",),
        out_lta_file=dict(argstr="--lta %s", min_ver="5.2.0",),
        out_reg_file=dict(argstr="--reg %s", genfile=True,),
        reg_frame=dict(argstr="--frame %d", xor=["reg_middle_frame"],),
        reg_middle_frame=dict(argstr="--mid-frame", xor=["reg_frame"],),
        registered_file=dict(argstr="--o %s",),
        source_file=dict(argstr="--mov %s", copyfile=False, mandatory=True,),
        spm_nifti=dict(argstr="--spm-nii",),
        subject_id=dict(argstr="--s %s", mandatory=True,),
        subjects_dir=dict(),
    )

    instance = BBRegister()
    if isinstance(instance.inputs, BBRegisterInputSpec6):
        input_map = input_map_6_0
    else:
        input_map = input_map_5_3

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(instance.inputs.traits()[key], metakey) == value


def test_BBRegister_outputs():
    output_map = dict(
        init_cost_file=dict(),
        min_cost_file=dict(),
        out_fsl_file=dict(),
        out_lta_file=dict(),
        out_reg_file=dict(),
        registered_file=dict(),
    )
    outputs = BBRegister.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..gtract import gtractFiberTracking


def test_gtractFiberTracking_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        branchingAngle=dict(
            argstr="--branchingAngle %f",
        ),
        branchingThreshold=dict(
            argstr="--branchingThreshold %f",
        ),
        curvatureThreshold=dict(
            argstr="--curvatureThreshold %f",
        ),
        endingSeedsLabel=dict(
            argstr="--endingSeedsLabel %d",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        guidedCurvatureThreshold=dict(
            argstr="--guidedCurvatureThreshold %f",
        ),
        inputAnisotropyVolume=dict(
            argstr="--inputAnisotropyVolume %s",
            extensions=None,
        ),
        inputEndingSeedsLabelMapVolume=dict(
            argstr="--inputEndingSeedsLabelMapVolume %s",
            extensions=None,
        ),
        inputStartingSeedsLabelMapVolume=dict(
            argstr="--inputStartingSeedsLabelMapVolume %s",
            extensions=None,
        ),
        inputTensorVolume=dict(
            argstr="--inputTensorVolume %s",
            extensions=None,
        ),
        inputTract=dict(
            argstr="--inputTract %s",
            extensions=None,
        ),
        maximumBranchPoints=dict(
            argstr="--maximumBranchPoints %d",
        ),
        maximumGuideDistance=dict(
            argstr="--maximumGuideDistance %f",
        ),
        maximumLength=dict(
            argstr="--maximumLength %f",
        ),
        minimumLength=dict(
            argstr="--minimumLength %f",
        ),
        numberOfThreads=dict(
            argstr="--numberOfThreads %d",
        ),
        outputTract=dict(
            argstr="--outputTract %s",
            hash_files=False,
        ),
        randomSeed=dict(
            argstr="--randomSeed %d",
        ),
        seedThreshold=dict(
            argstr="--seedThreshold %f",
        ),
        startingSeedsLabel=dict(
            argstr="--startingSeedsLabel %d",
        ),
        stepSize=dict(
            argstr="--stepSize %f",
        ),
        tendF=dict(
            argstr="--tendF %f",
        ),
        tendG=dict(
            argstr="--tendG %f",
        ),
        trackingMethod=dict(
            argstr="--trackingMethod %s",
        ),
        trackingThreshold=dict(
            argstr="--trackingThreshold %f",
        ),
        useLoopDetection=dict(
            argstr="--useLoopDetection ",
        ),
        useRandomWalk=dict(
            argstr="--useRandomWalk ",
        ),
        useTend=dict(
            argstr="--useTend ",
        ),
        writeXMLPolyDataFile=dict(
            argstr="--writeXMLPolyDataFile ",
        ),
    )
    inputs = gtractFiberTracking.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_gtractFiberTracking_outputs():
    output_map = dict(
        outputTract=dict(
            extensions=None,
        ),
    )
    outputs = gtractFiberTracking.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brains import FindCenterOfBrain


def test_FindCenterOfBrain_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        axis=dict(
            argstr="--axis %d",
        ),
        backgroundValue=dict(
            argstr="--backgroundValue %d",
        ),
        clippedImageMask=dict(
            argstr="--clippedImageMask %s",
            hash_files=False,
        ),
        closingSize=dict(
            argstr="--closingSize %d",
        ),
        debugAfterGridComputationsForegroundImage=dict(
            argstr="--debugAfterGridComputationsForegroundImage %s",
            hash_files=False,
        ),
        debugClippedImageMask=dict(
            argstr="--debugClippedImageMask %s",
            hash_files=False,
        ),
        debugDistanceImage=dict(
            argstr="--debugDistanceImage %s",
            hash_files=False,
        ),
        debugGridImage=dict(
            argstr="--debugGridImage %s",
            hash_files=False,
        ),
        debugTrimmedImage=dict(
            argstr="--debugTrimmedImage %s",
            hash_files=False,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        generateDebugImages=dict(
            argstr="--generateDebugImages ",
        ),
        headSizeEstimate=dict(
            argstr="--headSizeEstimate %f",
        ),
        headSizeLimit=dict(
            argstr="--headSizeLimit %f",
        ),
        imageMask=dict(
            argstr="--imageMask %s",
            extensions=None,
        ),
        inputVolume=dict(
            argstr="--inputVolume %s",
            extensions=None,
        ),
        maximize=dict(
            argstr="--maximize ",
        ),
        otsuPercentileThreshold=dict(
            argstr="--otsuPercentileThreshold %f",
        ),
    )
    inputs = FindCenterOfBrain.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_FindCenterOfBrain_outputs():
    output_map = dict(
        clippedImageMask=dict(
            extensions=None,
        ),
        debugAfterGridComputationsForegroundImage=dict(
            extensions=None,
        ),
        debugClippedImageMask=dict(
            extensions=None,
        ),
        debugDistanceImage=dict(
            extensions=None,
        ),
        debugGridImage=dict(
            extensions=None,
        ),
        debugTrimmedImage=dict(
            extensions=None,
        ),
    )
    outputs = FindCenterOfBrain.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import pytest

from ....testing import example_data
from ...niftyreg import get_custom_path

from ..dwi import FitDwi, DwiTool
from ...niftyreg.tests.test_regutils import no_nifty_tool


@pytest.mark.skipif(no_nifty_tool(cmd="fit_dwi"), reason="niftyfit is not installed")
def test_fit_dwi():
    """ Testing FitDwi interface."""
    # Create a node object
    fit_dwi = FitDwi()

    # Check if the command is properly defined
    cmd = get_custom_path("fit_dwi", env_dir="NIFTYFITDIR")
    assert fit_dwi.cmd == cmd

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        fit_dwi.run()

    # Assign some input data
    in_file = example_data("dwi.nii.gz")
    bval_file = example_data("bvals")
    bvec_file = example_data("bvecs")
    fit_dwi.inputs.source_file = in_file
    fit_dwi.inputs.bval_file = bval_file
    fit_dwi.inputs.bvec_file = bvec_file
    fit_dwi.inputs.dti_flag = True

    cmd_tmp = "{cmd} -source {in_file} -bval {bval} -bvec {bvec} -dti \
-error {error} -famap {fa} -mcmap {mc} -mcout {mcout} -mdmap {md} -nodiff \
{nodiff} -res {res} -rgbmap {rgb} -syn {syn} -tenmap2 {ten2}  -v1map {v1}"

    expected_cmd = cmd_tmp.format(
        cmd=cmd,
        in_file=in_file,
        bval=bval_file,
        bvec=bvec_file,
        error="dwi_error.nii.gz",
        fa="dwi_famap.nii.gz",
        mc="dwi_mcmap.nii.gz",
        md="dwi_mdmap.nii.gz",
        nodiff="dwi_no_diff.nii.gz",
        res="dwi_resmap.nii.gz",
        rgb="dwi_rgbmap.nii.gz",
        syn="dwi_syn.nii.gz",
        ten2="dwi_tenmap2.nii.gz",
        v1="dwi_v1map.nii.gz",
        mcout="dwi_mcout.txt",
    )

    assert fit_dwi.cmdline == expected_cmd


@pytest.mark.skipif(no_nifty_tool(cmd="dwi_tool"), reason="niftyfit is not installed")
def test_dwi_tool():
    """ Testing DwiTool interface."""
    # Create a node object
    dwi_tool = DwiTool()

    # Check if the command is properly defined
    cmd = get_custom_path("dwi_tool", env_dir="NIFTYFITDIR")
    assert dwi_tool.cmd == cmd

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        dwi_tool.run()

    # Assign some input data
    in_file = example_data("dwi.nii.gz")
    bval_file = example_data("bvals")
    bvec_file = example_data("bvecs")
    b0_file = example_data("b0.nii")
    mask_file = example_data("mask.nii.gz")
    dwi_tool.inputs.source_file = in_file
    dwi_tool.inputs.mask_file = mask_file
    dwi_tool.inputs.bval_file = bval_file
    dwi_tool.inputs.bvec_file = bvec_file
    dwi_tool.inputs.b0_file = b0_file
    dwi_tool.inputs.dti_flag = True

    cmd_tmp = "{cmd} -source {in_file} -bval {bval} -bvec {bvec} -b0 {b0} \
-mask {mask} -dti -famap {fa} -logdti2 {log} -mcmap {mc} -mdmap {md} \
-rgbmap {rgb} -syn {syn} -v1map {v1}"

    expected_cmd = cmd_tmp.format(
        cmd=cmd,
        in_file=in_file,
        bval=bval_file,
        bvec=bvec_file,
        b0=b0_file,
        mask=mask_file,
        fa="dwi_famap.nii.gz",
        log="dwi_logdti2.nii.gz",
        mc="dwi_mcmap.nii.gz",
        md="dwi_mdmap.nii.gz",
        rgb="dwi_rgbmap.nii.gz",
        syn="dwi_syn.nii.gz",
        v1="dwi_v1map.nii.gz",
    )

    assert dwi_tool.cmdline == expected_cmd



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Tests for workflow callbacks
"""
from time import sleep
import pytest
import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe


def func():
    return


def bad_func():
    raise Exception


class Status(object):
    def __init__(self):
        self.statuses = []

    def callback(self, node, status, result=None):
        self.statuses.append((node.name, status))


@pytest.mark.parametrize("plugin", ["Linear", "MultiProc", "LegacyMultiProc"])
def test_callback_normal(tmpdir, plugin):
    tmpdir.chdir()

    so = Status()
    wf = pe.Workflow(name="test", base_dir=tmpdir.strpath)
    f_node = pe.Node(
        niu.Function(function=func, input_names=[], output_names=[]), name="f_node"
    )
    wf.add_nodes([f_node])
    wf.config["execution"] = {"crashdump_dir": wf.base_dir, "poll_sleep_duration": 2}
    wf.run(plugin=plugin, plugin_args={"status_callback": so.callback})
    assert so.statuses == [("f_node", "start"), ("f_node", "end")]


@pytest.mark.parametrize("plugin", ["Linear", "MultiProc", "LegacyMultiProc"])
@pytest.mark.parametrize("stop_on_first_crash", [False, True])
def test_callback_exception(tmpdir, plugin, stop_on_first_crash):
    tmpdir.chdir()

    so = Status()
    wf = pe.Workflow(name="test", base_dir=tmpdir.strpath)
    f_node = pe.Node(
        niu.Function(function=bad_func, input_names=[], output_names=[]), name="f_node"
    )
    wf.add_nodes([f_node])
    wf.config["execution"] = {
        "crashdump_dir": wf.base_dir,
        "stop_on_first_crash": stop_on_first_crash,
        "poll_sleep_duration": 2,
    }
    with pytest.raises(Exception):
        wf.run(plugin=plugin, plugin_args={"status_callback": so.callback})

    sleep(0.5)  # Wait for callback to be called (python 2.7)
    assert so.statuses == [("f_node", "start"), ("f_node", "exception")]



# -*- coding: utf-8 -*-

import os.path as op
import numpy as np
import nibabel as nb
from distutils.version import LooseVersion

from ... import logging
from ..base import TraitedSpec, BaseInterfaceInputSpec, File, isdefined, traits
from .base import (
    DipyBaseInterface,
    HAVE_DIPY,
    dipy_version,
    dipy_to_nipype_interface,
    get_dipy_workflows,
)

IFLOGGER = logging.getLogger("nipype.interface")


if HAVE_DIPY and (
    LooseVersion("0.15") >= LooseVersion(dipy_version()) >= LooseVersion("0.16")
):
    try:
        from dipy.workflows.tracking import LocalFiberTrackingPAMFlow as DetTrackFlow
    except ImportError:  # different name in 0.15
        from dipy.workflows.tracking import DetTrackPAMFlow as DetTrackFlow

    DeterministicTracking = dipy_to_nipype_interface(
        "DeterministicTracking", DetTrackFlow
    )

if HAVE_DIPY and LooseVersion(dipy_version()) >= LooseVersion("0.15"):
    from dipy.workflows import segment, tracking

    l_wkflw = get_dipy_workflows(segment) + get_dipy_workflows(tracking)
    for name, obj in l_wkflw:
        new_name = name.replace("Flow", "")
        globals()[new_name] = dipy_to_nipype_interface(new_name, obj)
    del l_wkflw

else:
    IFLOGGER.info(
        "We advise you to upgrade DIPY version. This upgrade will"
        " open access to more function"
    )


class TrackDensityMapInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc="The input TrackVis track file")
    reference = File(
        exists=True, desc="A reference file to define RAS coordinates space"
    )
    points_space = traits.Enum(
        "rasmm", "voxel", None, usedefault=True, desc="coordinates of trk file"
    )
    voxel_dims = traits.List(
        traits.Float, minlen=3, maxlen=3, desc="The size of each voxel in mm."
    )
    data_dims = traits.List(
        traits.Int, minlen=3, maxlen=3, desc="The size of the image in voxels."
    )
    out_filename = File(
        "tdi.nii",
        usedefault=True,
        desc="The output filename for the tracks in TrackVis " "(.trk) format",
    )


class TrackDensityMapOutputSpec(TraitedSpec):
    out_file = File(exists=True)


class TrackDensityMap(DipyBaseInterface):
    """
    Creates a tract density image from a TrackVis track file using functions
    from dipy

    Example
    -------

    >>> import nipype.interfaces.dipy as dipy
    >>> trk2tdi = dipy.TrackDensityMap()
    >>> trk2tdi.inputs.in_file = 'converted.trk'
    >>> trk2tdi.run()                                   # doctest: +SKIP

    """

    input_spec = TrackDensityMapInputSpec
    output_spec = TrackDensityMapOutputSpec

    def _run_interface(self, runtime):
        from numpy import min_scalar_type
        from dipy.tracking.utils import density_map
        import nibabel.trackvis as nbt

        tracks, header = nbt.read(self.inputs.in_file)
        streams = ((ii[0]) for ii in tracks)

        if isdefined(self.inputs.reference):
            refnii = nb.load(self.inputs.reference)
            affine = refnii.affine
            data_dims = refnii.shape[:3]
            kwargs = dict(affine=affine)
        else:
            IFLOGGER.warning(
                "voxel_dims and data_dims are deprecated as of dipy "
                "0.7.1. Please use reference input instead"
            )

            if not isdefined(self.inputs.data_dims):
                data_dims = header["dim"]
            else:
                data_dims = self.inputs.data_dims
            if not isdefined(self.inputs.voxel_dims):
                voxel_size = header["voxel_size"]
            else:
                voxel_size = self.inputs.voxel_dims

            affine = header["vox_to_ras"]
            kwargs = dict(voxel_size=voxel_size)

        data = density_map(streams, data_dims, **kwargs)
        data = data.astype(min_scalar_type(data.max()))
        img = nb.Nifti1Image(data, affine)
        out_file = op.abspath(self.inputs.out_filename)
        nb.save(img, out_file)

        IFLOGGER.info(
            "Track density map saved as %s, size=%s, dimensions=%s",
            out_file,
            img.shape,
            img.header.get_zooms(),
        )

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_file"] = op.abspath(self.inputs.out_filename)
        return outputs


class StreamlineTractographyInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc=("input diffusion data"))
    in_model = File(exists=True, desc=("input f/d-ODF model extracted from."))
    tracking_mask = File(exists=True, desc=("input mask within which perform tracking"))
    seed_mask = File(exists=True, desc=("input mask within which perform seeding"))
    in_peaks = File(exists=True, desc=("peaks computed from the odf"))
    seed_coord = File(
        exists=True,
        desc=("file containing the list of seed voxel " "coordinates (N,3)"),
    )
    gfa_thresh = traits.Float(
        0.2,
        mandatory=True,
        usedefault=True,
        desc=("GFA threshold to compute tracking mask"),
    )
    peak_threshold = traits.Float(
        0.5,
        mandatory=True,
        usedefault=True,
        desc=("threshold to consider peaks from model"),
    )
    min_angle = traits.Float(
        25.0, mandatory=True, usedefault=True, desc=("minimum separation angle")
    )
    multiprocess = traits.Bool(
        True, mandatory=True, usedefault=True, desc=("use multiprocessing")
    )
    save_seeds = traits.Bool(
        False, mandatory=True, usedefault=True, desc=("save seeding voxels coordinates")
    )
    num_seeds = traits.Int(
        10000,
        mandatory=True,
        usedefault=True,
        desc=("desired number of tracks in tractography"),
    )
    out_prefix = traits.Str(desc=("output prefix for file names"))


class StreamlineTractographyOutputSpec(TraitedSpec):
    tracks = File(desc="TrackVis file containing extracted streamlines")
    gfa = File(
        desc=(
            "The resulting GFA (generalized FA) computed using the " "peaks of the ODF"
        )
    )
    odf_peaks = File(desc=("peaks computed from the odf"))
    out_seeds = File(
        desc=("file containing the (N,3) *voxel* coordinates used" " in seeding.")
    )


class StreamlineTractography(DipyBaseInterface):
    """
    Streamline tractography using EuDX [Garyfallidis12]_.

    .. [Garyfallidis12] Garyfallidis E., “Towards an accurate brain
      tractography”, PhD thesis, University of Cambridge, 2012

    Example
    -------

    >>> from nipype.interfaces import dipy as ndp
    >>> track = ndp.StreamlineTractography()
    >>> track.inputs.in_file = '4d_dwi.nii'
    >>> track.inputs.in_model = 'model.pklz'
    >>> track.inputs.tracking_mask = 'dilated_wm_mask.nii'
    >>> res = track.run() # doctest: +SKIP
    """

    input_spec = StreamlineTractographyInputSpec
    output_spec = StreamlineTractographyOutputSpec

    def _run_interface(self, runtime):
        from dipy.reconst.peaks import peaks_from_model
        from dipy.tracking.eudx import EuDX
        from dipy.data import get_sphere

        # import marshal as pickle
        import pickle as pickle
        import gzip

        if not (isdefined(self.inputs.in_model) or isdefined(self.inputs.in_peaks)):
            raise RuntimeError(
                ("At least one of in_model or in_peaks should " "be supplied")
            )

        img = nb.load(self.inputs.in_file)
        imref = nb.four_to_three(img)[0]
        affine = img.affine

        data = img.get_fdata(dtype=np.float32)
        hdr = imref.header.copy()
        hdr.set_data_dtype(np.float32)
        hdr["data_type"] = 16

        sphere = get_sphere("symmetric724")

        self._save_peaks = False
        if isdefined(self.inputs.in_peaks):
            IFLOGGER.info("Peaks file found, skipping ODF peaks search...")
            f = gzip.open(self.inputs.in_peaks, "rb")
            peaks = pickle.load(f)
            f.close()
        else:
            self._save_peaks = True
            IFLOGGER.info("Loading model and computing ODF peaks")
            f = gzip.open(self.inputs.in_model, "rb")
            odf_model = pickle.load(f)
            f.close()

            peaks = peaks_from_model(
                model=odf_model,
                data=data,
                sphere=sphere,
                relative_peak_threshold=self.inputs.peak_threshold,
                min_separation_angle=self.inputs.min_angle,
                parallel=self.inputs.multiprocess,
            )

            f = gzip.open(self._gen_filename("peaks", ext=".pklz"), "wb")
            pickle.dump(peaks, f, -1)
            f.close()

        hdr.set_data_shape(peaks.gfa.shape)
        nb.Nifti1Image(peaks.gfa.astype(np.float32), affine, hdr).to_filename(
            self._gen_filename("gfa")
        )

        IFLOGGER.info("Performing tractography")

        if isdefined(self.inputs.tracking_mask):
            msk = np.asanyarray(nb.load(self.inputs.tracking_mask).dataobj)
            msk[msk > 0] = 1
            msk[msk < 0] = 0
        else:
            msk = np.ones(imref.shape)

        gfa = peaks.gfa * msk
        seeds = self.inputs.num_seeds

        if isdefined(self.inputs.seed_coord):
            seeds = np.loadtxt(self.inputs.seed_coord)

        elif isdefined(self.inputs.seed_mask):
            seedmsk = np.asanyarray(nb.load(self.inputs.seed_mask).dataobj)
            assert seedmsk.shape == data.shape[:3]
            seedmsk[seedmsk > 0] = 1
            seedmsk[seedmsk < 1] = 0
            seedps = np.array(np.where(seedmsk == 1), dtype=np.float32).T
            vseeds = seedps.shape[0]
            nsperv = (seeds // vseeds) + 1
            IFLOGGER.info(
                "Seed mask is provided (%d voxels inside "
                "mask), computing seeds (%d seeds/voxel).",
                vseeds,
                nsperv,
            )
            if nsperv > 1:
                IFLOGGER.info(
                    "Needed %d seeds per selected voxel (total %d).", nsperv, vseeds
                )
                seedps = np.vstack(np.array([seedps] * nsperv))
                voxcoord = seedps + np.random.uniform(-1, 1, size=seedps.shape)
                nseeds = voxcoord.shape[0]
                seeds = affine.dot(np.vstack((voxcoord.T, np.ones((1, nseeds)))))[
                    :3, :
                ].T

                if self.inputs.save_seeds:
                    np.savetxt(self._gen_filename("seeds", ext=".txt"), seeds)

        if isdefined(self.inputs.tracking_mask):
            tmask = msk
            a_low = 0.1
        else:
            tmask = gfa
            a_low = self.inputs.gfa_thresh

        eu = EuDX(
            tmask,
            peaks.peak_indices[..., 0],
            seeds=seeds,
            affine=affine,
            odf_vertices=sphere.vertices,
            a_low=a_low,
        )

        ss_mm = [np.array(s) for s in eu]

        trkfilev = nb.trackvis.TrackvisFile(
            [(s, None, None) for s in ss_mm], points_space="rasmm", affine=np.eye(4)
        )
        trkfilev.to_file(self._gen_filename("tracked", ext=".trk"))
        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["tracks"] = self._gen_filename("tracked", ext=".trk")
        outputs["gfa"] = self._gen_filename("gfa")
        if self._save_peaks:
            outputs["odf_peaks"] = self._gen_filename("peaks", ext=".pklz")
        if self.inputs.save_seeds:
            if isdefined(self.inputs.seed_coord):
                outputs["out_seeds"] = self.inputs.seed_coord
            else:
                outputs["out_seeds"] = self._gen_filename("seeds", ext=".txt")

        return outputs

    def _gen_filename(self, name, ext=None):
        fname, fext = op.splitext(op.basename(self.inputs.in_file))
        if fext == ".gz":
            fname, fext2 = op.splitext(fname)
            fext = fext2 + fext

        if not isdefined(self.inputs.out_prefix):
            out_prefix = op.abspath(fname)
        else:
            out_prefix = self.inputs.out_prefix

        if ext is None:
            ext = fext

        return out_prefix + "_" + name + ext



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import ImageStats


def test_ImageStats_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_files=dict(
            argstr="-images %s",
            mandatory=True,
            position=-1,
        ),
        out_type=dict(
            argstr="-outputdatatype %s",
            usedefault=True,
        ),
        output_root=dict(
            argstr="-outputroot %s",
            extensions=None,
            mandatory=True,
        ),
        stat=dict(
            argstr="-stat %s",
            mandatory=True,
            units="NA",
        ),
    )
    inputs = ImageStats.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ImageStats_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = ImageStats.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
import os
import numpy as np
from numpy import ones, kron, mean, eye, hstack, dot, tile
from numpy.linalg import pinv
import nibabel as nb
from ..interfaces.base import (
    BaseInterfaceInputSpec,
    TraitedSpec,
    BaseInterface,
    traits,
    File,
)


class ICCInputSpec(BaseInterfaceInputSpec):
    subjects_sessions = traits.List(
        traits.List(File(exists=True)),
        desc="n subjects m sessions 3D stat files",
        mandatory=True,
    )
    mask = File(exists=True, mandatory=True)


class ICCOutputSpec(TraitedSpec):
    icc_map = File(exists=True)
    session_var_map = File(exists=True, desc="variance between sessions")
    subject_var_map = File(exists=True, desc="variance between subjects")


class ICC(BaseInterface):
    """
    Calculates Interclass Correlation Coefficient (3,1) as defined in
    P. E. Shrout & Joseph L. Fleiss (1979). "Intraclass Correlations: Uses in
    Assessing Rater Reliability". Psychological Bulletin 86 (2): 420-428. This
    particular implementation is aimed at relaibility (test-retest) studies.
    """

    input_spec = ICCInputSpec
    output_spec = ICCOutputSpec

    def _run_interface(self, runtime):
        maskdata = nb.load(self.inputs.mask).get_fdata()
        maskdata = np.logical_not(np.logical_or(maskdata == 0, np.isnan(maskdata)))

        session_datas = [
            [nb.load(fname).get_fdata()[maskdata].reshape(-1, 1) for fname in sessions]
            for sessions in self.inputs.subjects_sessions
        ]
        list_of_sessions = [np.dstack(session_data) for session_data in session_datas]
        all_data = np.hstack(list_of_sessions)
        icc = np.zeros(session_datas[0][0].shape)
        session_F = np.zeros(session_datas[0][0].shape)
        session_var = np.zeros(session_datas[0][0].shape)
        subject_var = np.zeros(session_datas[0][0].shape)

        for x in range(icc.shape[0]):
            Y = all_data[x, :, :]
            icc[x], subject_var[x], session_var[x], session_F[x], _, _ = ICC_rep_anova(
                Y
            )

        nim = nb.load(self.inputs.subjects_sessions[0][0])
        new_data = np.zeros(nim.shape)
        new_data[maskdata] = icc.reshape(-1)
        new_img = nb.Nifti1Image(new_data, nim.affine, nim.header)
        nb.save(new_img, "icc_map.nii")

        new_data = np.zeros(nim.shape)
        new_data[maskdata] = session_var.reshape(-1)
        new_img = nb.Nifti1Image(new_data, nim.affine, nim.header)
        nb.save(new_img, "session_var_map.nii")

        new_data = np.zeros(nim.shape)
        new_data[maskdata] = subject_var.reshape(-1)
        new_img = nb.Nifti1Image(new_data, nim.affine, nim.header)
        nb.save(new_img, "subject_var_map.nii")

        return runtime

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["icc_map"] = os.path.abspath("icc_map.nii")
        outputs["session_var_map"] = os.path.abspath("session_var_map.nii")
        outputs["subject_var_map"] = os.path.abspath("subject_var_map.nii")
        return outputs


def ICC_rep_anova(Y):
    """
    the data Y are entered as a 'table' ie subjects are in rows and repeated
    measures in columns

    One Sample Repeated measure ANOVA

    Y = XB + E with X = [FaTor / Subjects]
    """

    [nb_subjects, nb_conditions] = Y.shape
    dfc = nb_conditions - 1
    dfe = (nb_subjects - 1) * dfc
    dfr = nb_subjects - 1

    # Compute the repeated measure effect
    # ------------------------------------

    # Sum Square Total
    mean_Y = mean(Y)
    SST = ((Y - mean_Y) ** 2).sum()

    # create the design matrix for the different levels
    x = kron(eye(nb_conditions), ones((nb_subjects, 1)))  # sessions
    x0 = tile(eye(nb_subjects), (nb_conditions, 1))  # subjects
    X = hstack([x, x0])

    # Sum Square Error
    predicted_Y = dot(dot(dot(X, pinv(dot(X.T, X))), X.T), Y.flatten("F"))
    residuals = Y.flatten("F") - predicted_Y
    SSE = (residuals ** 2).sum()

    residuals.shape = Y.shape

    MSE = SSE / dfe

    # Sum square session effect - between colums/sessions
    SSC = ((mean(Y, 0) - mean_Y) ** 2).sum() * nb_subjects
    MSC = SSC / dfc / nb_subjects

    session_effect_F = MSC / MSE

    # Sum Square subject effect - between rows/subjects
    SSR = SST - SSC - SSE
    MSR = SSR / dfr

    # ICC(3,1) = (mean square subjeT - mean square error) /
    #            (mean square subjeT + (k-1)*-mean square error)
    ICC = (MSR - MSE) / (MSR + dfc * MSE)

    e_var = MSE  # variance of error
    r_var = (MSR - MSE) / nb_conditions  # variance between subjects

    return ICC, r_var, e_var, session_effect_F, dfc, dfe



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import ApplyVolTransform


def test_ApplyVolTransform_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        fs_target=dict(
            argstr="--fstarg",
            mandatory=True,
            requires=["reg_file"],
            xor=("target_file", "tal", "fs_target"),
        ),
        fsl_reg_file=dict(
            argstr="--fsl %s",
            extensions=None,
            mandatory=True,
            xor=(
                "reg_file",
                "lta_file",
                "lta_inv_file",
                "fsl_reg_file",
                "xfm_reg_file",
                "reg_header",
                "mni_152_reg",
                "subject",
            ),
        ),
        interp=dict(
            argstr="--interp %s",
        ),
        inverse=dict(
            argstr="--inv",
        ),
        invert_morph=dict(
            argstr="--inv-morph",
            requires=["m3z_file"],
        ),
        lta_file=dict(
            argstr="--lta %s",
            extensions=None,
            mandatory=True,
            xor=(
                "reg_file",
                "lta_file",
                "lta_inv_file",
                "fsl_reg_file",
                "xfm_reg_file",
                "reg_header",
                "mni_152_reg",
                "subject",
            ),
        ),
        lta_inv_file=dict(
            argstr="--lta-inv %s",
            extensions=None,
            mandatory=True,
            xor=(
                "reg_file",
                "lta_file",
                "lta_inv_file",
                "fsl_reg_file",
                "xfm_reg_file",
                "reg_header",
                "mni_152_reg",
                "subject",
            ),
        ),
        m3z_file=dict(
            argstr="--m3z %s",
            extensions=None,
        ),
        mni_152_reg=dict(
            argstr="--regheader",
            mandatory=True,
            xor=(
                "reg_file",
                "lta_file",
                "lta_inv_file",
                "fsl_reg_file",
                "xfm_reg_file",
                "reg_header",
                "mni_152_reg",
                "subject",
            ),
        ),
        no_ded_m3z_path=dict(
            argstr="--noDefM3zPath",
            requires=["m3z_file"],
        ),
        no_resample=dict(
            argstr="--no-resample",
        ),
        reg_file=dict(
            argstr="--reg %s",
            extensions=None,
            mandatory=True,
            xor=(
                "reg_file",
                "lta_file",
                "lta_inv_file",
                "fsl_reg_file",
                "xfm_reg_file",
                "reg_header",
                "mni_152_reg",
                "subject",
            ),
        ),
        reg_header=dict(
            argstr="--regheader",
            mandatory=True,
            xor=(
                "reg_file",
                "lta_file",
                "lta_inv_file",
                "fsl_reg_file",
                "xfm_reg_file",
                "reg_header",
                "mni_152_reg",
                "subject",
            ),
        ),
        source_file=dict(
            argstr="--mov %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        subject=dict(
            argstr="--s %s",
            mandatory=True,
            xor=(
                "reg_file",
                "lta_file",
                "lta_inv_file",
                "fsl_reg_file",
                "xfm_reg_file",
                "reg_header",
                "mni_152_reg",
                "subject",
            ),
        ),
        subjects_dir=dict(),
        tal=dict(
            argstr="--tal",
            mandatory=True,
            xor=("target_file", "tal", "fs_target"),
        ),
        tal_resolution=dict(
            argstr="--talres %.10f",
        ),
        target_file=dict(
            argstr="--targ %s",
            extensions=None,
            mandatory=True,
            xor=("target_file", "tal", "fs_target"),
        ),
        transformed_file=dict(
            argstr="--o %s",
            extensions=None,
            genfile=True,
        ),
        xfm_reg_file=dict(
            argstr="--xfm %s",
            extensions=None,
            mandatory=True,
            xor=(
                "reg_file",
                "lta_file",
                "lta_inv_file",
                "fsl_reg_file",
                "xfm_reg_file",
                "reg_header",
                "mni_152_reg",
                "subject",
            ),
        ),
    )
    inputs = ApplyVolTransform.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_ApplyVolTransform_outputs():
    output_map = dict(
        transformed_file=dict(
            extensions=None,
        ),
    )
    outputs = ApplyVolTransform.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..utils import MotionOutliers


def test_MotionOutliers_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        dummy=dict(
            argstr="--dummy=%d",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        in_file=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
        ),
        mask=dict(
            argstr="-m %s",
            extensions=None,
        ),
        metric=dict(
            argstr="--%s",
        ),
        no_motion_correction=dict(
            argstr="--nomoco",
        ),
        out_file=dict(
            argstr="-o %s",
            extensions=None,
            hash_files=False,
            keep_extension=True,
            name_source="in_file",
            name_template="%s_outliers.txt",
        ),
        out_metric_plot=dict(
            argstr="-p %s",
            extensions=None,
            hash_files=False,
            keep_extension=True,
            name_source="in_file",
            name_template="%s_metrics.png",
        ),
        out_metric_values=dict(
            argstr="-s %s",
            extensions=None,
            hash_files=False,
            keep_extension=True,
            name_source="in_file",
            name_template="%s_metrics.txt",
        ),
        output_type=dict(),
        threshold=dict(
            argstr="--thresh=%g",
        ),
    )
    inputs = MotionOutliers.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_MotionOutliers_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
        out_metric_plot=dict(
            extensions=None,
        ),
        out_metric_values=dict(
            extensions=None,
        ),
    )
    outputs = MotionOutliers.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import os
import pytest

from ....utils.filemanip import which
from ....testing import example_data
from .. import (
    get_custom_path,
    RegAverage,
    RegResample,
    RegJacobian,
    RegTools,
    RegMeasure,
    RegTransform,
)


def no_nifty_tool(cmd=None):
    return which(cmd) is None


@pytest.mark.skipif(
    no_nifty_tool(cmd="reg_resample"),
    reason="niftyreg is not installed. reg_resample not found.",
)
def test_reg_resample_res():
    """ tests for reg_resample interface """
    # Create a reg_resample object
    nr_resample = RegResample()

    # Check if the command is properly defined
    assert nr_resample.cmd == get_custom_path("reg_resample")

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_resample.run()

    # Resample res
    ref_file = example_data("im1.nii")
    flo_file = example_data("im2.nii")
    trans_file = example_data("warpfield.nii")
    nr_resample.inputs.ref_file = ref_file
    nr_resample.inputs.flo_file = flo_file
    nr_resample.inputs.trans_file = trans_file
    nr_resample.inputs.inter_val = "LIN"
    nr_resample.inputs.omp_core_val = 4

    cmd_tmp = "{cmd} -flo {flo} -inter 1 -omp 4 -ref {ref} -trans {trans} \
-res {res}"

    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_resample"),
        flo=flo_file,
        ref=ref_file,
        trans=trans_file,
        res="im2_res.nii.gz",
    )

    assert nr_resample.cmdline == expected_cmd

    # test_reg_resample_blank()
    nr_resample_2 = RegResample(type="blank", inter_val="LIN", omp_core_val=4)
    ref_file = example_data("im1.nii")
    flo_file = example_data("im2.nii")
    trans_file = example_data("warpfield.nii")
    nr_resample_2.inputs.ref_file = ref_file
    nr_resample_2.inputs.flo_file = flo_file
    nr_resample_2.inputs.trans_file = trans_file

    cmd_tmp = "{cmd} -flo {flo} -inter 1 -omp 4 -ref {ref} -trans {trans} \
-blank {blank}"

    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_resample"),
        flo=flo_file,
        ref=ref_file,
        trans=trans_file,
        blank="im2_blank.nii.gz",
    )

    assert nr_resample_2.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd="reg_jacobian"),
    reason="niftyreg is not installed. reg_jacobian not found.",
)
def test_reg_jacobian_jac():
    """ Test interface for RegJacobian """
    # Create a reg_jacobian object
    nr_jacobian = RegJacobian()

    # Check if the command is properly defined
    assert nr_jacobian.cmd == get_custom_path("reg_jacobian")

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_jacobian.run()

    # Test Reg Jacobian: jac
    ref_file = example_data("im1.nii")
    trans_file = example_data("warpfield.nii")
    nr_jacobian.inputs.ref_file = ref_file
    nr_jacobian.inputs.trans_file = trans_file
    nr_jacobian.inputs.omp_core_val = 4

    cmd_tmp = "{cmd} -omp 4 -ref {ref} -trans {trans} -jac {jac}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_jacobian"),
        ref=ref_file,
        trans=trans_file,
        jac="warpfield_jac.nii.gz",
    )

    assert nr_jacobian.cmdline == expected_cmd

    # Test Reg Jacobian: jac m
    nr_jacobian_2 = RegJacobian(type="jacM", omp_core_val=4)
    ref_file = example_data("im1.nii")
    trans_file = example_data("warpfield.nii")
    nr_jacobian_2.inputs.ref_file = ref_file
    nr_jacobian_2.inputs.trans_file = trans_file

    cmd_tmp = "{cmd} -omp 4 -ref {ref} -trans {trans} -jacM {jac}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_jacobian"),
        ref=ref_file,
        trans=trans_file,
        jac="warpfield_jacM.nii.gz",
    )

    assert nr_jacobian_2.cmdline == expected_cmd

    # Test Reg Jacobian: jac l
    nr_jacobian_3 = RegJacobian(type="jacL", omp_core_val=4)
    ref_file = example_data("im1.nii")
    trans_file = example_data("warpfield.nii")
    nr_jacobian_3.inputs.ref_file = ref_file
    nr_jacobian_3.inputs.trans_file = trans_file

    cmd_tmp = "{cmd} -omp 4 -ref {ref} -trans {trans} -jacL {jac}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_jacobian"),
        ref=ref_file,
        trans=trans_file,
        jac="warpfield_jacL.nii.gz",
    )

    assert nr_jacobian_3.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd="reg_tools"),
    reason="niftyreg is not installed. reg_tools not found.",
)
def test_reg_tools_mul():
    """ tests for reg_tools interface """
    # Create a reg_tools object
    nr_tools = RegTools()

    # Check if the command is properly defined
    assert nr_tools.cmd == get_custom_path("reg_tools")

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_tools.run()

    # Test reg_tools: mul
    in_file = example_data("im1.nii")
    nr_tools.inputs.in_file = in_file
    nr_tools.inputs.mul_val = 4
    nr_tools.inputs.omp_core_val = 4

    cmd_tmp = "{cmd} -in {in_file} -mul 4.0 -omp 4 -out {out_file}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_tools"), in_file=in_file, out_file="im1_tools.nii.gz"
    )

    assert nr_tools.cmdline == expected_cmd

    # Test reg_tools: iso
    nr_tools_2 = RegTools(iso_flag=True, omp_core_val=4)
    in_file = example_data("im1.nii")
    nr_tools_2.inputs.in_file = in_file

    cmd_tmp = "{cmd} -in {in_file} -iso -omp 4 -out {out_file}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_tools"), in_file=in_file, out_file="im1_tools.nii.gz"
    )

    assert nr_tools_2.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd="reg_average"),
    reason="niftyreg is not installed. reg_average not found.",
)
def test_reg_average():
    """ tests for reg_average interface """
    # Create a reg_average object
    nr_average = RegAverage()

    # Check if the command is properly defined
    assert nr_average.cmd == get_custom_path("reg_average")

    # Average niis
    one_file = example_data("im1.nii")
    two_file = example_data("im2.nii")
    three_file = example_data("im3.nii")
    nr_average.inputs.avg_files = [one_file, two_file, three_file]
    nr_average.inputs.omp_core_val = 1
    generated_cmd = nr_average.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), "reg_average_cmd")
    with open(reg_average_cmd, "rb") as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = "%s %s -avg %s %s %s -omp 1" % (
        get_custom_path("reg_average"),
        os.path.join(os.getcwd(), "avg_out.nii.gz"),
        one_file,
        two_file,
        three_file,
    )

    assert argv.decode("utf-8") == expected_argv

    # Test command line with text file
    expected_cmd = "%s --cmd_file %s" % (
        get_custom_path("reg_average"),
        reg_average_cmd,
    )

    assert generated_cmd == expected_cmd

    # Test Reg Average: average txt
    nr_average_2 = RegAverage()
    one_file = example_data("TransformParameters.0.txt")
    two_file = example_data("ants_Affine.txt")
    three_file = example_data("elastix.txt")
    nr_average_2.inputs.avg_files = [one_file, two_file, three_file]
    nr_average_2.inputs.omp_core_val = 1
    generated_cmd = nr_average_2.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), "reg_average_cmd")
    with open(reg_average_cmd, "rb") as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = "%s %s -avg %s %s %s -omp 1" % (
        get_custom_path("reg_average"),
        os.path.join(os.getcwd(), "avg_out.txt"),
        one_file,
        two_file,
        three_file,
    )

    assert argv.decode("utf-8") == expected_argv

    # Test Reg Average: average list
    nr_average_3 = RegAverage()
    one_file = example_data("TransformParameters.0.txt")
    two_file = example_data("ants_Affine.txt")
    three_file = example_data("elastix.txt")
    nr_average_3.inputs.avg_lts_files = [one_file, two_file, three_file]
    nr_average_3.inputs.omp_core_val = 1
    generated_cmd = nr_average_3.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), "reg_average_cmd")
    with open(reg_average_cmd, "rb") as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = "%s %s -avg_lts %s %s %s -omp 1" % (
        get_custom_path("reg_average"),
        os.path.join(os.getcwd(), "avg_out.txt"),
        one_file,
        two_file,
        three_file,
    )

    assert argv.decode("utf-8") == expected_argv

    # Test Reg Average: average ref
    nr_average_4 = RegAverage()
    ref_file = example_data("anatomical.nii")
    one_file = example_data("im1.nii")
    two_file = example_data("im2.nii")
    three_file = example_data("im3.nii")
    trans1_file = example_data("roi01.nii")
    trans2_file = example_data("roi02.nii")
    trans3_file = example_data("roi03.nii")
    nr_average_4.inputs.warp_files = [
        trans1_file,
        one_file,
        trans2_file,
        two_file,
        trans3_file,
        three_file,
    ]
    nr_average_4.inputs.avg_ref_file = ref_file
    nr_average_4.inputs.omp_core_val = 1
    generated_cmd = nr_average_4.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), "reg_average_cmd")
    with open(reg_average_cmd, "rb") as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = "%s %s -avg_tran %s -omp 1 %s %s %s %s %s %s" % (
        get_custom_path("reg_average"),
        os.path.join(os.getcwd(), "avg_out.nii.gz"),
        ref_file,
        trans1_file,
        one_file,
        trans2_file,
        two_file,
        trans3_file,
        three_file,
    )

    assert argv.decode("utf-8") == expected_argv

    # Test Reg Average: demean3
    nr_average_5 = RegAverage()
    ref_file = example_data("anatomical.nii")
    one_file = example_data("im1.nii")
    two_file = example_data("im2.nii")
    three_file = example_data("im3.nii")
    aff1_file = example_data("TransformParameters.0.txt")
    aff2_file = example_data("ants_Affine.txt")
    aff3_file = example_data("elastix.txt")
    trans1_file = example_data("roi01.nii")
    trans2_file = example_data("roi02.nii")
    trans3_file = example_data("roi03.nii")
    nr_average_5.inputs.warp_files = [
        aff1_file,
        trans1_file,
        one_file,
        aff2_file,
        trans2_file,
        two_file,
        aff3_file,
        trans3_file,
        three_file,
    ]
    nr_average_5.inputs.demean3_ref_file = ref_file
    nr_average_5.inputs.omp_core_val = 1
    generated_cmd = nr_average_5.cmdline

    # Read the reg_average_cmd
    reg_average_cmd = os.path.join(os.getcwd(), "reg_average_cmd")
    with open(reg_average_cmd, "rb") as f_obj:
        argv = f_obj.read()
    os.remove(reg_average_cmd)

    expected_argv = "%s %s -demean3 %s -omp 1 %s %s %s %s %s %s %s %s %s" % (
        get_custom_path("reg_average"),
        os.path.join(os.getcwd(), "avg_out.nii.gz"),
        ref_file,
        aff1_file,
        trans1_file,
        one_file,
        aff2_file,
        trans2_file,
        two_file,
        aff3_file,
        trans3_file,
        three_file,
    )

    assert argv.decode("utf-8") == expected_argv


@pytest.mark.skipif(
    no_nifty_tool(cmd="reg_transform"),
    reason="niftyreg is not installed. reg_transform not found.",
)
def test_reg_transform_def():
    """ tests for reg_transform interface """
    # Create a reg_transform object
    nr_transform = RegTransform()

    # Check if the command is properly defined
    assert nr_transform.cmd == get_custom_path("reg_transform")

    # Assign some input data
    trans_file = example_data("warpfield.nii")
    nr_transform.inputs.def_input = trans_file
    nr_transform.inputs.omp_core_val = 4

    cmd_tmp = "{cmd} -omp 4 -def {trans_file} {out_file}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_transform"),
        trans_file=trans_file,
        out_file=os.path.join(os.getcwd(), "warpfield_trans.nii.gz"),
    )

    assert nr_transform.cmdline == expected_cmd

    # Test reg_transform: def ref
    nr_transform_2 = RegTransform(omp_core_val=4)
    ref_file = example_data("im1.nii")
    trans_file = example_data("warpfield.nii")
    nr_transform_2.inputs.ref1_file = ref_file
    nr_transform_2.inputs.def_input = trans_file

    cmd_tmp = "{cmd} -ref {ref_file} -omp 4 -def {trans_file} {out_file}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_transform"),
        ref_file=ref_file,
        trans_file=trans_file,
        out_file=os.path.join(os.getcwd(), "warpfield_trans.nii.gz"),
    )

    assert nr_transform_2.cmdline == expected_cmd

    # Test reg_transform: comp nii
    nr_transform_3 = RegTransform(omp_core_val=4)
    ref_file = example_data("im1.nii")
    trans_file = example_data("warpfield.nii")
    trans2_file = example_data("anatomical.nii")
    nr_transform_3.inputs.ref1_file = ref_file
    nr_transform_3.inputs.comp_input2 = trans2_file
    nr_transform_3.inputs.comp_input = trans_file

    cmd_tmp = "{cmd} -ref {ref_file} -omp 4 -comp {trans1} {trans2} {out_file}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_transform"),
        ref_file=ref_file,
        trans1=trans_file,
        trans2=trans2_file,
        out_file=os.path.join(os.getcwd(), "warpfield_trans.nii.gz"),
    )

    assert nr_transform_3.cmdline == expected_cmd

    # Test reg_transform: comp txt
    nr_transform_4 = RegTransform(omp_core_val=4)
    aff1_file = example_data("ants_Affine.txt")
    aff2_file = example_data("elastix.txt")
    nr_transform_4.inputs.comp_input2 = aff2_file
    nr_transform_4.inputs.comp_input = aff1_file

    cmd_tmp = "{cmd} -omp 4 -comp {aff1} {aff2} {out_file}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_transform"),
        aff1=aff1_file,
        aff2=aff2_file,
        out_file=os.path.join(os.getcwd(), "ants_Affine_trans.txt"),
    )

    assert nr_transform_4.cmdline == expected_cmd

    # Test reg_transform: comp
    nr_transform_5 = RegTransform(omp_core_val=4)
    trans_file = example_data("warpfield.nii")
    aff_file = example_data("elastix.txt")
    nr_transform_5.inputs.comp_input2 = trans_file
    nr_transform_5.inputs.comp_input = aff_file

    cmd_tmp = "{cmd} -omp 4 -comp {aff} {trans} {out_file}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_transform"),
        aff=aff_file,
        trans=trans_file,
        out_file=os.path.join(os.getcwd(), "elastix_trans.nii.gz"),
    )

    assert nr_transform_5.cmdline == expected_cmd

    # Test reg_transform: flirt
    nr_transform_6 = RegTransform(omp_core_val=4)
    aff_file = example_data("elastix.txt")
    ref_file = example_data("im1.nii")
    in_file = example_data("im2.nii")
    nr_transform_6.inputs.flirt_2_nr_input = (aff_file, ref_file, in_file)

    cmd_tmp = "{cmd} -omp 4 -flirtAff2NR {aff} {ref} {in_file} {out_file}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_transform"),
        aff=aff_file,
        ref=ref_file,
        in_file=in_file,
        out_file=os.path.join(os.getcwd(), "elastix_trans.txt"),
    )

    assert nr_transform_6.cmdline == expected_cmd


@pytest.mark.skipif(
    no_nifty_tool(cmd="reg_measure"),
    reason="niftyreg is not installed. reg_measure not found.",
)
def test_reg_measure():
    """ tests for reg_measure interface """
    # Create a reg_measure object
    nr_measure = RegMeasure()

    # Check if the command is properly defined
    assert nr_measure.cmd == get_custom_path("reg_measure")

    # test raising error with mandatory args absent
    with pytest.raises(ValueError):
        nr_measure.run()

    # Assign some input data
    ref_file = example_data("im1.nii")
    flo_file = example_data("im2.nii")
    nr_measure.inputs.ref_file = ref_file
    nr_measure.inputs.flo_file = flo_file
    nr_measure.inputs.measure_type = "lncc"
    nr_measure.inputs.omp_core_val = 4

    cmd_tmp = "{cmd} -flo {flo} -lncc -omp 4 -out {out} -ref {ref}"
    expected_cmd = cmd_tmp.format(
        cmd=get_custom_path("reg_measure"),
        flo=flo_file,
        out="im2_lncc.txt",
        ref=ref_file,
    )

    assert nr_measure.cmdline == expected_cmd



# -*- coding: utf-8 -*-
"""The bru2nii module provides basic functions for dicom conversion
"""

import os
from .base import (
    CommandLine,
    CommandLineInputSpec,
    traits,
    TraitedSpec,
    isdefined,
    File,
    Directory,
)


class Bru2InputSpec(CommandLineInputSpec):
    input_dir = Directory(
        desc="Input Directory", exists=True, mandatory=True, position=-1, argstr="%s"
    )
    actual_size = traits.Bool(
        argstr="-a",
        desc="Keep actual size - otherwise x10 scale so animals match human.",
    )
    force_conversion = traits.Bool(
        argstr="-f",
        desc="Force conversion of localizers images (multiple slice " "orientations).",
    )
    compress = traits.Bool(argstr="-z", desc='gz compress images (".nii.gz").')
    append_protocol_name = traits.Bool(
        argstr="-p", desc="Append protocol name to output filename."
    )
    output_filename = traits.Str(
        argstr="-o %s",
        desc='Output filename (".nii" will be appended, or ".nii.gz" if the "-z" compress option is selected)',
        genfile=True,
    )


class Bru2OutputSpec(TraitedSpec):
    nii_file = File(exists=True)


class Bru2(CommandLine):
    """Uses bru2nii's Bru2 to convert Bruker files

    Examples
    ========

    >>> from nipype.interfaces.bru2nii import Bru2
    >>> converter = Bru2()
    >>> converter.inputs.input_dir = "brukerdir"
    >>> converter.cmdline  # doctest: +ELLIPSIS
    'Bru2 -o .../data/brukerdir brukerdir'
    """

    input_spec = Bru2InputSpec
    output_spec = Bru2OutputSpec
    _cmd = "Bru2"

    def _list_outputs(self):
        outputs = self._outputs().get()
        if isdefined(self.inputs.output_filename):
            output_filename1 = os.path.abspath(self.inputs.output_filename)
        else:
            output_filename1 = self._gen_filename("output_filename")
        if self.inputs.compress:
            outputs["nii_file"] = output_filename1 + ".nii.gz"
        else:
            outputs["nii_file"] = output_filename1 + ".nii"
        return outputs

    def _gen_filename(self, name):
        if name == "output_filename":
            outfile = os.path.join(
                os.getcwd(), os.path.basename(os.path.normpath(self.inputs.input_dir))
            )
            return outfile



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Provides interfaces to various longitudinal commands provided by freesurfer
"""

import os
import os.path

from ... import logging
from ...utils.filemanip import split_filename, copyfile

from .base import (
    FSCommand,
    FSTraitedSpec,
    FSScriptCommand,
    FSScriptOutputSpec,
    FSCommandOpenMP,
    FSTraitedSpecOpenMP,
)
from ..base import isdefined, TraitedSpec, File, traits, Directory

__docformat__ = "restructuredtext"
iflogger = logging.getLogger("nipype.interface")


class MPRtoMNI305InputSpec(FSTraitedSpec):
    # environment variables, required
    # usedefault=True is hack for on_trait_change in __init__
    reference_dir = Directory(
        "", exists=True, mandatory=True, usedefault=True, desc="TODO"
    )
    target = traits.String("", mandatory=True, usedefault=True, desc="input atlas file")
    # required
    in_file = File(
        argstr="%s", usedefault=True, desc="the input file prefix for MPRtoMNI305"
    )


class MPRtoMNI305OutputSpec(FSScriptOutputSpec):
    out_file = File(
        exists=False, desc="The output file '<in_file>_to_<target>_t4_vox2vox.txt'"
    )


class MPRtoMNI305(FSScriptCommand):
    """
    For complete details, see FreeSurfer documentation

    Examples
    --------
    >>> from nipype.interfaces.freesurfer import MPRtoMNI305, Info
    >>> mprtomni305 = MPRtoMNI305()
    >>> mprtomni305.inputs.target = 'structural.nii'
    >>> mprtomni305.inputs.reference_dir = '.' # doctest: +SKIP
    >>> mprtomni305.cmdline # doctest: +SKIP
    'mpr2mni305 output'
    >>> mprtomni305.inputs.out_file = 'struct_out' # doctest: +SKIP
    >>> mprtomni305.cmdline # doctest: +SKIP
    'mpr2mni305 struct_out' # doctest: +SKIP
    >>> mprtomni305.inputs.environ['REFDIR'] == os.path.join(Info.home(), 'average') # doctest: +SKIP
    True
    >>> mprtomni305.inputs.environ['MPR2MNI305_TARGET'] # doctest: +SKIP
    'structural'
    >>> mprtomni305.run() # doctest: +SKIP

    """

    _cmd = "mpr2mni305"
    input_spec = MPRtoMNI305InputSpec
    output_spec = MPRtoMNI305OutputSpec

    def __init__(self, **inputs):
        super(MPRtoMNI305, self).__init__(**inputs)
        self.inputs.on_trait_change(self._environ_update, "target")
        self.inputs.on_trait_change(self._environ_update, "reference_dir")

    def _format_arg(self, opt, spec, val):
        if opt in ["target", "reference_dir"]:
            return ""
        elif opt == "in_file":
            _, retval, ext = split_filename(val)
            # Need to copy file to working cache directory!
            copyfile(
                val, os.path.abspath(retval + ext), copy=True, hashmethod="content"
            )
            return retval
        return super(MPRtoMNI305, self)._format_arg(opt, spec, val)

    def _environ_update(self):
        # refdir = os.path.join(Info.home(), val)
        refdir = self.inputs.reference_dir
        target = self.inputs.target
        self.inputs.environ["MPR2MNI305_TARGET"] = target
        self.inputs.environ["REFDIR"] = refdir

    def _get_fname(self, fname):
        return split_filename(fname)[1]

    def _list_outputs(self):
        outputs = super(MPRtoMNI305, self)._list_outputs()
        fullname = "_".join(
            [
                self._get_fname(self.inputs.in_file),
                "to",
                self.inputs.target,
                "t4",
                "vox2vox.txt",
            ]
        )
        outputs["out_file"] = os.path.abspath(fullname)
        return outputs


class RegisterAVItoTalairachInputSpec(FSTraitedSpec):
    in_file = File(
        argstr="%s", exists=True, mandatory=True, position=0, desc="The input file"
    )
    target = File(
        argstr="%s", exists=True, mandatory=True, position=1, desc="The target file"
    )
    vox2vox = File(
        argstr="%s", exists=True, mandatory=True, position=2, desc="The vox2vox file"
    )
    out_file = File(
        "talairach.auto.xfm",
        usedefault=True,
        argstr="%s",
        position=3,
        desc="The transform output",
    )


class RegisterAVItoTalairachOutputSpec(FSScriptOutputSpec):
    out_file = File(exists=False, desc="The output file for RegisterAVItoTalairach")


class RegisterAVItoTalairach(FSScriptCommand):
    """
    converts the vox2vox from talairach_avi to a talairach.xfm file

    This is a script that converts the vox2vox from talairach_avi to a
    talairach.xfm file. It is meant to replace the following cmd line:

    tkregister2_cmdl \
        --mov $InVol \
        --targ $FREESURFER_HOME/average/mni305.cor.mgz \
        --xfmout ${XFM} \
        --vox2vox talsrcimg_to_${target}_t4_vox2vox.txt \
        --noedit \
        --reg talsrcimg.reg.tmp.dat
    set targ = $FREESURFER_HOME/average/mni305.cor.mgz
    set subject = mgh-02407836-v2
    set InVol = $SUBJECTS_DIR/$subject/mri/orig.mgz
    set vox2vox = $SUBJECTS_DIR/$subject/mri/transforms/talsrcimg_to_711-2C_as_mni_average_305_t4_vox2vox.txt

    Examples
    ========

    >>> from nipype.interfaces.freesurfer import RegisterAVItoTalairach
    >>> register = RegisterAVItoTalairach()
    >>> register.inputs.in_file = 'structural.mgz'                         # doctest: +SKIP
    >>> register.inputs.target = 'mni305.cor.mgz'                          # doctest: +SKIP
    >>> register.inputs.vox2vox = 'talsrcimg_to_structural_t4_vox2vox.txt' # doctest: +SKIP
    >>> register.cmdline                                                   # doctest: +SKIP
    'avi2talxfm structural.mgz mni305.cor.mgz talsrcimg_to_structural_t4_vox2vox.txt talairach.auto.xfm'

    >>> register.run() # doctest: +SKIP
    """

    _cmd = "avi2talxfm"
    input_spec = RegisterAVItoTalairachInputSpec
    output_spec = RegisterAVItoTalairachOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        return outputs


class EMRegisterInputSpec(FSTraitedSpecOpenMP):
    # required
    in_file = File(
        argstr="%s", exists=True, mandatory=True, position=-3, desc="in brain volume"
    )
    template = File(
        argstr="%s", exists=True, mandatory=True, position=-2, desc="template gca"
    )
    out_file = File(
        argstr="%s",
        exists=False,
        name_source=["in_file"],
        name_template="%s_transform.lta",
        hash_files=False,
        keep_extension=False,
        position=-1,
        desc="output transform",
    )
    # optional
    skull = traits.Bool(argstr="-skull", desc="align to atlas containing skull (uns=5)")
    mask = File(argstr="-mask %s", exists=True, desc="use volume as a mask")
    nbrspacing = traits.Int(
        argstr="-uns %d",
        desc="align to atlas containing skull setting unknown_nbr_spacing = nbrspacing",
    )
    transform = File(argstr="-t %s", exists=True, desc="Previously computed transform")


class EMRegisterOutputSpec(TraitedSpec):
    out_file = File(exists=False, desc="output transform")


class EMRegister(FSCommandOpenMP):
    """ This program creates a tranform in lta format

    Examples
    ========
    >>> from nipype.interfaces.freesurfer import EMRegister
    >>> register = EMRegister()
    >>> register.inputs.in_file = 'norm.mgz'
    >>> register.inputs.template = 'aseg.mgz'
    >>> register.inputs.out_file = 'norm_transform.lta'
    >>> register.inputs.skull = True
    >>> register.inputs.nbrspacing = 9
    >>> register.cmdline
    'mri_em_register -uns 9 -skull norm.mgz aseg.mgz norm_transform.lta'
    """

    _cmd = "mri_em_register"
    input_spec = EMRegisterInputSpec
    output_spec = EMRegisterOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        return outputs


class RegisterInputSpec(FSTraitedSpec):
    # required
    in_surf = File(
        argstr="%s",
        exists=True,
        mandatory=True,
        position=-3,
        copyfile=True,
        desc="Surface to register, often {hemi}.sphere",
    )
    target = File(
        argstr="%s",
        exists=True,
        mandatory=True,
        position=-2,
        desc="The data to register to. In normal recon-all usage, "
        + "this is a template file for average surface.",
    )
    in_sulc = File(
        exists=True,
        mandatory=True,
        copyfile=True,
        desc="Undocumented mandatory input file ${SUBJECTS_DIR}/surf/{hemisphere}.sulc ",
    )
    out_file = File(
        argstr="%s",
        exists=False,
        position=-1,
        genfile=True,
        desc="Output surface file to capture registration",
    )
    # optional
    curv = traits.Bool(
        argstr="-curv",
        requires=["in_smoothwm"],
        desc="Use smoothwm curvature for final alignment",
    )
    in_smoothwm = File(
        exists=True,
        copyfile=True,
        desc="Undocumented input file ${SUBJECTS_DIR}/surf/{hemisphere}.smoothwm ",
    )


class RegisterOutputSpec(TraitedSpec):
    out_file = File(exists=False, desc="Output surface file to capture registration")


class Register(FSCommand):
    """ This program registers a surface to an average surface template.

    Examples
    ========
    >>> from nipype.interfaces.freesurfer import Register
    >>> register = Register()
    >>> register.inputs.in_surf = 'lh.pial'
    >>> register.inputs.in_smoothwm = 'lh.pial'
    >>> register.inputs.in_sulc = 'lh.pial'
    >>> register.inputs.target = 'aseg.mgz'
    >>> register.inputs.out_file = 'lh.pial.reg'
    >>> register.inputs.curv = True
    >>> register.cmdline
    'mris_register -curv lh.pial aseg.mgz lh.pial.reg'
    """

    _cmd = "mris_register"
    input_spec = RegisterInputSpec
    output_spec = RegisterOutputSpec

    def _format_arg(self, opt, spec, val):
        if opt == "curv":
            return spec.argstr
        return super(Register, self)._format_arg(opt, spec, val)

    def _gen_filename(self, name):
        if name == "out_file":
            return self._list_outputs()[name]
        return None

    def _list_outputs(self):
        outputs = self.output_spec().get()
        if isdefined(self.inputs.out_file):
            outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        else:
            outputs["out_file"] = os.path.abspath(self.inputs.in_surf) + ".reg"
        return outputs


class PaintInputSpec(FSTraitedSpec):
    # required
    in_surf = File(
        argstr="%s",
        exists=True,
        mandatory=True,
        position=-2,
        desc="Surface file with grid (vertices) onto which the "
        + "template data is to be sampled or 'painted'",
    )
    template = File(
        argstr="%s", exists=True, mandatory=True, position=-3, desc="Template file"
    )
    # optional
    template_param = traits.Int(desc="Frame number of the input template")
    averages = traits.Int(argstr="-a %d", desc="Average curvature patterns")
    out_file = File(
        argstr="%s",
        exists=False,
        position=-1,
        name_template="%s.avg_curv",
        hash_files=False,
        name_source=["in_surf"],
        keep_extension=False,
        desc="File containing a surface-worth of per-vertex values, "
        + "saved in 'curvature' format.",
    )


class PaintOutputSpec(TraitedSpec):
    out_file = File(
        exists=False,
        desc="File containing a surface-worth of per-vertex values, saved in 'curvature' format.",
    )


class Paint(FSCommand):
    """
    This program is useful for extracting one of the arrays ("a variable")
    from a surface-registration template file. The output is a file
    containing a surface-worth of per-vertex values, saved in "curvature"
    format. Because the template data is sampled to a particular surface
    mesh, this conjures the idea of "painting to a surface".

    Examples
    ========
    >>> from nipype.interfaces.freesurfer import Paint
    >>> paint = Paint()
    >>> paint.inputs.in_surf = 'lh.pial'
    >>> paint.inputs.template = 'aseg.mgz'
    >>> paint.inputs.averages = 5
    >>> paint.inputs.out_file = 'lh.avg_curv'
    >>> paint.cmdline
    'mrisp_paint -a 5 aseg.mgz lh.pial lh.avg_curv'
    """

    _cmd = "mrisp_paint"
    input_spec = PaintInputSpec
    output_spec = PaintOutputSpec

    def _format_arg(self, opt, spec, val):
        if opt == "template":
            if isdefined(self.inputs.template_param):
                return spec.argstr % (val + "#" + str(self.inputs.template_param))
        return super(Paint, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        return outputs


class MRICoregInputSpec(FSTraitedSpec):
    source_file = File(
        argstr="--mov %s",
        desc="source file to be registered",
        mandatory=True,
        copyfile=False,
    )
    reference_file = File(
        argstr="--ref %s",
        desc="reference (target) file",
        mandatory=True,
        copyfile=False,
        xor=["subject_id"],
    )
    out_lta_file = traits.Either(
        True,
        File,
        argstr="--lta %s",
        default=True,
        usedefault=True,
        desc="output registration file (LTA format)",
    )
    out_reg_file = traits.Either(
        True, File, argstr="--regdat %s", desc="output registration file (REG format)"
    )
    out_params_file = traits.Either(
        True, File, argstr="--params %s", desc="output parameters file"
    )

    subjects_dir = Directory(
        exists=True, argstr="--sd %s", desc="FreeSurfer SUBJECTS_DIR"
    )
    subject_id = traits.Str(
        argstr="--s %s",
        position=1,
        mandatory=True,
        xor=["reference_file"],
        requires=["subjects_dir"],
        desc="freesurfer subject ID (implies ``reference_mask == "
        "aparc+aseg.mgz`` unless otherwise specified)",
    )
    dof = traits.Enum(
        6, 9, 12, argstr="--dof %d", desc="number of transform degrees of freedom"
    )
    reference_mask = traits.Either(
        False,
        traits.Str,
        argstr="--ref-mask %s",
        position=2,
        desc="mask reference volume with given mask, or None if ``False``",
    )
    source_mask = traits.Str(
        argstr="--mov-mask", desc="mask source file with given mask"
    )
    num_threads = traits.Int(argstr="--threads %d", desc="number of OpenMP threads")
    no_coord_dithering = traits.Bool(
        argstr="--no-coord-dither", desc="turn off coordinate dithering"
    )
    no_intensity_dithering = traits.Bool(
        argstr="--no-intensity-dither", desc="turn off intensity dithering"
    )
    sep = traits.List(
        argstr="--sep %s...",
        minlen=1,
        maxlen=2,
        desc="set spatial scales, in voxels (default [2, 4])",
    )
    initial_translation = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        argstr="--trans %g %g %g",
        desc="initial translation in mm (implies no_cras0)",
    )
    initial_rotation = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        argstr="--rot %g %g %g",
        desc="initial rotation in degrees",
    )
    initial_scale = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        argstr="--scale %g %g %g",
        desc="initial scale",
    )
    initial_shear = traits.Tuple(
        traits.Float,
        traits.Float,
        traits.Float,
        argstr="--shear %g %g %g",
        desc="initial shear (Hxy, Hxz, Hyz)",
    )
    no_cras0 = traits.Bool(
        argstr="--no-cras0",
        desc="do not set translation parameters to align "
        "centers of source and reference files",
    )
    max_iters = traits.Range(
        low=1, argstr="--nitersmax %d", desc="maximum iterations (default: 4)"
    )
    ftol = traits.Float(
        argstr="--ftol %e", desc="floating-point tolerance (default=1e-7)"
    )
    linmintol = traits.Float(argstr="--linmintol %e")
    saturation_threshold = traits.Range(
        low=0.0,
        high=100.0,
        argstr="--sat %g",
        desc="saturation threshold (default=9.999)",
    )
    conform_reference = traits.Bool(
        argstr="--conf-ref", desc="conform reference without rescaling"
    )
    no_brute_force = traits.Bool(argstr="--no-bf", desc="do not brute force search")
    brute_force_limit = traits.Float(
        argstr="--bf-lim %g",
        xor=["no_brute_force"],
        desc="constrain brute force search to +/- lim",
    )
    brute_force_samples = traits.Int(
        argstr="--bf-nsamp %d",
        xor=["no_brute_force"],
        desc="number of samples in brute force search",
    )
    no_smooth = traits.Bool(
        argstr="--no-smooth",
        desc="do not apply smoothing to either reference or source file",
    )
    ref_fwhm = traits.Float(
        argstr="--ref-fwhm", desc="apply smoothing to reference file"
    )
    source_oob = traits.Bool(
        argstr="--mov-oob", desc="count source voxels that are out-of-bounds as 0"
    )
    # Skipping mat2par


class MRICoregOutputSpec(TraitedSpec):
    out_reg_file = File(exists=True, desc="output registration file")
    out_lta_file = File(exists=True, desc="output LTA-style registration file")
    out_params_file = File(exists=True, desc="output parameters file")


class MRICoreg(FSCommand):
    """ This program registers one volume to another

    mri_coreg is a C reimplementation of spm_coreg in FreeSurfer

    Examples
    ========
    >>> from nipype.interfaces.freesurfer import MRICoreg
    >>> coreg = MRICoreg()
    >>> coreg.inputs.source_file = 'moving1.nii'
    >>> coreg.inputs.reference_file = 'fixed1.nii'
    >>> coreg.inputs.subjects_dir = '.'
    >>> coreg.cmdline # doctest: +ELLIPSIS
    'mri_coreg --lta .../registration.lta --ref fixed1.nii --mov moving1.nii --sd .'

    If passing a subject ID, the reference mask may be disabled:

    >>> coreg = MRICoreg()
    >>> coreg.inputs.source_file = 'moving1.nii'
    >>> coreg.inputs.subjects_dir = '.'
    >>> coreg.inputs.subject_id = 'fsaverage'
    >>> coreg.inputs.reference_mask = False
    >>> coreg.cmdline # doctest: +ELLIPSIS
    'mri_coreg --s fsaverage --no-ref-mask --lta .../registration.lta --mov moving1.nii --sd .'

    Spatial scales may be specified as a list of one or two separations:

    >>> coreg.inputs.sep = [4]
    >>> coreg.cmdline # doctest: +ELLIPSIS
    'mri_coreg --s fsaverage --no-ref-mask --lta .../registration.lta --sep 4 --mov moving1.nii --sd .'

    >>> coreg.inputs.sep = [4, 5]
    >>> coreg.cmdline # doctest: +ELLIPSIS
    'mri_coreg --s fsaverage --no-ref-mask --lta .../registration.lta --sep 4 --sep 5 --mov moving1.nii --sd .'
    """

    _cmd = "mri_coreg"
    input_spec = MRICoregInputSpec
    output_spec = MRICoregOutputSpec

    def _format_arg(self, opt, spec, val):
        if opt in ("out_reg_file", "out_lta_file", "out_params_file") and val is True:
            val = self._list_outputs()[opt]
        elif opt == "reference_mask" and val is False:
            return "--no-ref-mask"
        return super(MRICoreg, self)._format_arg(opt, spec, val)

    def _list_outputs(self):
        outputs = self.output_spec().get()

        out_lta_file = self.inputs.out_lta_file
        if isdefined(out_lta_file):
            if out_lta_file is True:
                out_lta_file = "registration.lta"
            outputs["out_lta_file"] = os.path.abspath(out_lta_file)

        out_reg_file = self.inputs.out_reg_file
        if isdefined(out_reg_file):
            if out_reg_file is True:
                out_reg_file = "registration.dat"
            outputs["out_reg_file"] = os.path.abspath(out_reg_file)

        out_params_file = self.inputs.out_params_file
        if isdefined(out_params_file):
            if out_params_file is True:
                out_params_file = "registration.par"
            outputs["out_params_file"] = os.path.abspath(out_params_file)

        return outputs



# -*- coding: utf-8 -*-
"""NIPY is a python project for analysis of structural and functional neuroimaging data."""
from .model import FitGLM, EstimateContrast
from .preprocess import ComputeMask, SpaceTimeRealigner
from .utils import Similarity



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Experimental Slicer wrapper - Work in progress."""
import os
import warnings
import xml.dom.minidom

from .base import (
    CommandLine,
    CommandLineInputSpec,
    DynamicTraitedSpec,
    traits,
    Undefined,
    File,
    isdefined,
)


class SlicerCommandLineInputSpec(DynamicTraitedSpec, CommandLineInputSpec):
    module = traits.Str(desc="name of the Slicer command line module you want to use")


class SlicerCommandLine(CommandLine):
    """Experimental Slicer wrapper. Work in progress."""

    _cmd = "Slicer3"
    input_spec = SlicerCommandLineInputSpec
    output_spec = DynamicTraitedSpec

    def _grab_xml(self, module):
        cmd = CommandLine(
            command="Slicer3", resource_monitor=False, args="--launch %s --xml" % module
        )
        ret = cmd.run()
        if ret.runtime.returncode == 0:
            return xml.dom.minidom.parseString(ret.runtime.stdout)
        else:
            raise Exception(cmd.cmdline + " failed:\n%s" % ret.runtime.stderr)

    def _outputs(self):
        base = super(SlicerCommandLine, self)._outputs()
        undefined_output_traits = {}
        for key in [
            node.getElementsByTagName("name")[0].firstChild.nodeValue
            for node in self._outputs_nodes
        ]:
            base.add_trait(key, File(exists=True))
            undefined_output_traits[key] = Undefined

        base.trait_set(trait_change_notify=False, **undefined_output_traits)
        return base

    def __init__(self, module, **inputs):
        warnings.warn("slicer is Not fully implemented", RuntimeWarning)
        super(SlicerCommandLine, self).__init__(
            command="Slicer3 --launch %s " % module, name=module, **inputs
        )
        dom = self._grab_xml(module)
        self._outputs_filenames = {}

        self._outputs_nodes = []

        undefined_traits = {}

        for paramGroup in dom.getElementsByTagName("parameters"):
            for param in paramGroup.childNodes:
                if param.nodeName in ["label", "description", "#text", "#comment"]:
                    continue
                traitsParams = {}

                name = param.getElementsByTagName("name")[0].firstChild.nodeValue

                longFlagNode = param.getElementsByTagName("longflag")
                if longFlagNode:
                    traitsParams["argstr"] = (
                        "--" + longFlagNode[0].firstChild.nodeValue + " "
                    )
                else:
                    traitsParams["argstr"] = "--" + name + " "

                argsDict = {
                    "file": "%s",
                    "integer": "%d",
                    "double": "%f",
                    "float": "%f",
                    "image": "%s",
                    "transform": "%s",
                    "boolean": "",
                    "string-enumeration": "%s",
                    "string": "%s",
                }

                if param.nodeName.endswith("-vector"):
                    traitsParams["argstr"] += argsDict[param.nodeName[:-7]]
                else:
                    traitsParams["argstr"] += argsDict[param.nodeName]

                index = param.getElementsByTagName("index")
                if index:
                    traitsParams["position"] = index[0].firstChild.nodeValue

                desc = param.getElementsByTagName("description")
                if index:
                    traitsParams["desc"] = desc[0].firstChild.nodeValue

                name = param.getElementsByTagName("name")[0].firstChild.nodeValue

                typesDict = {
                    "integer": traits.Int,
                    "double": traits.Float,
                    "float": traits.Float,
                    "image": File,
                    "transform": File,
                    "boolean": traits.Bool,
                    "string": traits.Str,
                    "file": File,
                }

                if param.nodeName == "string-enumeration":
                    type = traits.Enum
                    values = [
                        el.firstChild.nodeValue
                        for el in param.getElementsByTagName("element")
                    ]
                elif param.nodeName.endswith("-vector"):
                    type = traits.List
                    values = [typesDict[param.nodeName[:-7]]]
                    traitsParams["sep"] = ","
                else:
                    values = []
                    type = typesDict[param.nodeName]

                if (
                    param.nodeName in ["file", "directory", "image", "transform"]
                    and param.getElementsByTagName("channel")[0].firstChild.nodeValue
                    == "output"
                ):
                    self.inputs.add_trait(
                        name, traits.Either(traits.Bool, File, **traitsParams)
                    )
                    undefined_traits[name] = Undefined

                    # traitsParams["exists"] = True
                    self._outputs_filenames[name] = self._gen_filename_from_param(param)
                    # undefined_output_traits[name] = Undefined
                    # self._outputs().add_trait(name, File(*values, **traitsParams))
                    self._outputs_nodes.append(param)
                else:
                    if param.nodeName in ["file", "directory", "image", "transform"]:
                        traitsParams["exists"] = True
                    self.inputs.add_trait(name, type(*values, **traitsParams))
                    undefined_traits[name] = Undefined

        self.inputs.trait_set(trait_change_notify=False, **undefined_traits)
        for name in list(undefined_traits.keys()):
            _ = getattr(self.inputs, name)
        # self._outputs().trait_set(trait_change_notify=False, **undefined_output_traits)

    def _gen_filename(self, name):
        if name in self._outputs_filenames:
            return os.path.join(os.getcwd(), self._outputs_filenames[name])
        return None

    def _gen_filename_from_param(self, param):
        base = param.getElementsByTagName("name")[0].firstChild.nodeValue
        fileExtensions = param.getAttribute("fileExtensions")
        if fileExtensions:
            ext = fileExtensions
        else:
            ext = {"image": ".nii", "transform": ".txt", "file": ""}[param.nodeName]
        return base + ext

    def _list_outputs(self):
        outputs = self.output_spec().get()
        for output_node in self._outputs_nodes:
            name = output_node.getElementsByTagName("name")[0].firstChild.nodeValue
            outputs[name] = getattr(self.inputs, name)
            if isdefined(outputs[name]) and isinstance(outputs[name], bool):
                if outputs[name]:
                    outputs[name] = self._gen_filename(name)
                else:
                    outputs[name] = Undefined
        return outputs

    def _format_arg(self, name, spec, value):
        if name in [
            output_node.getElementsByTagName("name")[0].firstChild.nodeValue
            for output_node in self._outputs_nodes
        ]:
            if isinstance(value, bool):
                fname = self._gen_filename(name)
            else:
                fname = value
            return spec.argstr % fname
        return super(SlicerCommandLine, self)._format_arg(name, spec, value)


#    test = SlicerCommandLine(module="BRAINSFit")
#    test.inputs.fixedVolume = "/home/filo/workspace/fmri_tumour/data/pilot1/10_co_COR_3D_IR_PREP.nii"
#    test.inputs.movingVolume = "/home/filo/workspace/fmri_tumour/data/pilot1/2_line_bisection.nii"
#    test.inputs.outputTransform = True
#    test.inputs.transformType = ["Affine"]
#    print test.cmdline
#    print test.inputs
#    print test._outputs()
#    ret = test.run()

#    test = SlicerCommandLine(name="BRAINSResample")
#    test.inputs.referenceVolume = "/home/filo/workspace/fmri_tumour/data/pilot1/10_co_COR_3D_IR_PREP.nii"
#    test.inputs.inputVolume = "/home/filo/workspace/fmri_tumour/data/pilot1/2_line_bisection.nii"
#    test.inputs.outputVolume = True
#    test.inputs.warpTransform = "/home/filo/workspace/nipype/nipype/interfaces/outputTransform.mat"
#    print test.cmdline
#    ret = test.run()
#    print ret.runtime.stderr
#    print ret.runtime.returncode



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..odf import HARDIMat


def test_HARDIMat_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        bvals=dict(
            extensions=None,
            mandatory=True,
        ),
        bvecs=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        image_info=dict(
            argstr="-info %s",
            extensions=None,
        ),
        image_orientation_vectors=dict(
            argstr="-iop %f",
        ),
        oblique_correction=dict(
            argstr="-oc",
        ),
        odf_file=dict(
            argstr="-odf %s",
            extensions=None,
        ),
        order=dict(
            argstr="-order %s",
        ),
        out_file=dict(
            argstr="%s",
            extensions=None,
            position=2,
            usedefault=True,
        ),
        reference_file=dict(
            argstr="-ref %s",
            extensions=None,
        ),
    )
    inputs = HARDIMat.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_HARDIMat_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = HARDIMat.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-

import os

from ...utils.filemanip import split_filename
from ..base import (
    CommandLineInputSpec,
    CommandLine,
    traits,
    TraitedSpec,
    File,
    Directory,
    StdOutCommandLine,
    StdOutCommandLineInputSpec,
    isdefined,
    InputMultiPath,
)


class DTIFitInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=1,
        desc="voxel-order data filename",
    )

    bgmask = File(
        argstr="-bgmask %s",
        exists=True,
        desc=(
            "Provides the name of a file containing a background mask computed using, "
            "for example, FSL bet2 program. The mask file contains zero in background "
            "voxels and non-zero in foreground."
        ),
    )

    scheme_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=2,
        desc="Camino scheme file (b values / vectors, see camino.fsl2scheme)",
    )

    non_linear = traits.Bool(
        argstr="-nonlinear",
        position=3,
        desc="Use non-linear fitting instead of the default linear regression "
        "to the log measurements. ",
    )


class DTIFitOutputSpec(TraitedSpec):
    tensor_fitted = File(exists=True, desc="path/name of 4D volume in voxel order")


class DTIFit(StdOutCommandLine):
    """
    Reads diffusion MRI data, acquired using the acquisition scheme detailed in the scheme file,
    from the data file.

    Use non-linear fitting instead of the default linear regression to the log measurements.
    The data file stores the diffusion MRI data in voxel order with the measurements stored
    in big-endian format and ordered as in the scheme file.
    The default input data type is four-byte float.
    The default output data type is eight-byte double.
    See modelfit and camino for the format of the data file and scheme file.
    The program fits the diffusion tensor to each voxel and outputs the results,
    in voxel order and as big-endian eight-byte doubles, to the standard output.
    The program outputs eight values in each voxel:
    [exit code, ln(S(0)), D_xx, D_xy, D_xz, D_yy, D_yz, D_zz].
    An exit code of zero indicates no problems.
    For a list of other exit codes, see modelfit(1).
    The entry S(0) is an estimate of the signal at q=0.

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> fit = cmon.DTIFit()
    >>> fit.inputs.scheme_file = 'A.scheme'
    >>> fit.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> fit.run()                  # doctest: +SKIP

    """

    _cmd = "dtfit"
    input_spec = DTIFitInputSpec
    output_spec = DTIFitOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["tensor_fitted"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_DT.Bdouble"


class DTMetricInputSpec(CommandLineInputSpec):
    eigen_data = File(
        exists=True,
        argstr="-inputfile %s",
        mandatory=True,
        desc="voxel-order data filename",
    )

    metric = traits.Enum(
        "fa",
        "md",
        "rd",
        "l1",
        "l2",
        "l3",
        "tr",
        "ra",
        "2dfa",
        "cl",
        "cp",
        "cs",
        argstr="-stat %s",
        mandatory=True,
        desc="Specifies the metric to compute.",
    )

    inputdatatype = traits.Enum(
        "double",
        "float",
        "long",
        "int",
        "short",
        "char",
        argstr="-inputdatatype %s",
        usedefault=True,
        desc="Specifies the data type of the input data.",
    )

    outputdatatype = traits.Enum(
        "double",
        "float",
        "long",
        "int",
        "short",
        "char",
        argstr="-outputdatatype %s",
        usedefault=True,
        desc="Specifies the data type of the output data.",
    )

    data_header = File(
        argstr="-header %s",
        exists=True,
        desc=(
            "A Nifti .nii or .nii.gz file containing the header information. "
            "Usually this will be the header of the raw data file from which "
            "the diffusion tensors were reconstructed."
        ),
    )

    outputfile = File(
        argstr="-outputfile %s",
        genfile=True,
        desc=(
            "Output name. Output will be a .nii.gz file if data_header is provided and"
            "in voxel order with outputdatatype datatype (default: double) otherwise."
        ),
    )


class DTMetricOutputSpec(TraitedSpec):
    metric_stats = File(
        exists=True, desc="Diffusion Tensor statistics of the chosen metric"
    )


class DTMetric(CommandLine):
    """
    Computes tensor metric statistics based on the eigenvalues l1 >= l2 >= l3
    typically obtained from ComputeEigensystem.

    The full list of statistics is:

     - <cl> = (l1 - l2) / l1 , a measure of linearity
     - <cp> = (l2 - l3) / l1 , a measure of planarity
     - <cs> = l3 / l1 , a measure of isotropy
       with: cl + cp + cs = 1
     - <l1> = first eigenvalue
     - <l2> = second eigenvalue
     - <l3> = third eigenvalue
     - <tr> = l1 + l2 + l3
     - <md> = tr / 3
     - <rd> = (l2 + l3) / 2
     - <fa> = fractional anisotropy. (Basser et al, J Magn Reson B 1996)
     - <ra> = relative anisotropy (Basser et al, J Magn Reson B 1996)
     - <2dfa> = 2D FA of the two minor eigenvalues l2 and l3
       i.e. sqrt( 2 * [(l2 - <l>)^2 + (l3 - <l>)^2] / (l2^2 + l3^2) )
       with: <l> = (l2 + l3) / 2


    Example
    -------
    Compute the CP planar metric as float data type.

    >>> import nipype.interfaces.camino as cam
    >>> dtmetric = cam.DTMetric()
    >>> dtmetric.inputs.eigen_data = 'dteig.Bdouble'
    >>> dtmetric.inputs.metric = 'cp'
    >>> dtmetric.inputs.outputdatatype = 'float'
    >>> dtmetric.run()                  # doctest: +SKIP

    """

    _cmd = "dtshape"
    input_spec = DTMetricInputSpec
    output_spec = DTMetricOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["metric_stats"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        return self._gen_outputfile()

    def _gen_outputfile(self):
        outputfile = self.inputs.outputfile
        if not isdefined(outputfile):
            outputfile = self._gen_filename("outputfile")
        return outputfile

    def _gen_filename(self, name):
        if name == "outputfile":
            _, name, _ = split_filename(self.inputs.eigen_data)
            metric = self.inputs.metric
            datatype = self.inputs.outputdatatype
            if isdefined(self.inputs.data_header):
                filename = name + "_" + metric + ".nii.gz"
            else:
                filename = name + "_" + metric + ".B" + datatype
        return filename


class ModelFitInputSpec(StdOutCommandLineInputSpec):
    def _gen_model_options():  # @NoSelf
        """
        Generate all possible permutations of < multi - tensor > < single - tensor > options
        """

        single_tensor = ["dt", "restore", "algdt", "nldt_pos", "nldt", "ldt_wtd"]
        multi_tensor = [
            "cylcyl",
            "cylcyl_eq",
            "pospos",
            "pospos_eq",
            "poscyl",
            "poscyl_eq",
            "cylcylcyl",
            "cylcylcyl_eq",
            "pospospos",
            "pospospos_eq",
            "posposcyl",
            "posposcyl_eq",
            "poscylcyl",
            "poscylcyl_eq",
        ]
        other = ["adc", "ball_stick"]

        model_list = single_tensor
        model_list.extend(other)
        model_list.extend(
            [multi + " " + single for multi in multi_tensor for single in single_tensor]
        )
        return model_list

    model = traits.Enum(
        _gen_model_options(),
        argstr="-model %s",
        mandatory=True,
        desc="Specifies the model to be fit to the data.",
    )

    in_file = File(
        exists=True,
        argstr="-inputfile %s",
        mandatory=True,
        desc="voxel-order data filename",
    )

    inputdatatype = traits.Enum(
        "float",
        "char",
        "short",
        "int",
        "long",
        "double",
        argstr="-inputdatatype %s",
        desc="Specifies the data type of the input file. "
        "The input file must have BIG-ENDIAN ordering. "
        "By default, the input type is ``float``.",
    )

    scheme_file = File(
        exists=True,
        argstr="-schemefile %s",
        mandatory=True,
        desc="Camino scheme file (b values / vectors, see camino.fsl2scheme)",
    )

    outputfile = File(argstr="-outputfile %s", desc="Filename of the output file.")

    outlier = File(
        argstr="-outliermap %s",
        exists=True,
        desc="Specifies the name of the file to contain the outlier map generated by "
        "the RESTORE algorithm.",
    )

    noisemap = File(
        argstr="-noisemap %s",
        exists=True,
        desc="Specifies the name of the file to contain the estimated noise variance on the "
        "diffusion-weighted signal, generated by a weighted tensor fit. "
        "The data type of this file is big-endian double.",
    )

    residualmap = File(
        argstr="-residualmap %s",
        exists=True,
        desc="Specifies the name of the file to contain the weighted residual errors after "
        "computing a weighted linear tensor fit. "
        "One value is produced per measurement, in voxel order. "
        "The data type of this file is big-endian double. "
        "Images of the residuals for each measurement can be extracted with shredder.",
    )

    sigma = traits.Float(
        argstr="-sigma %G",
        desc="Specifies the standard deviation of the noise in the data. "
        "Required by the RESTORE algorithm.",
    )

    bgthresh = traits.Float(
        argstr="-bgthresh %G",
        desc="Sets a threshold on the average q=0 measurement to separate "
        "foreground and background. The program does not process background voxels, "
        "but outputs the same number of values in background voxels and foreground voxels. "
        "Each value is zero in background voxels apart from the exit code which is -1.",
    )

    bgmask = File(
        argstr="-bgmask %s",
        exists=True,
        desc="Provides the name of a file containing a background mask computed using, "
        "for example, FSL's bet2 program. The mask file contains zero in background voxels "
        "and non-zero in foreground.",
    )

    cfthresh = traits.Float(
        argstr="-csfthresh %G",
        desc="Sets a threshold on the average q=0 measurement to determine which voxels "
        "are CSF. This program does not treat CSF voxels any different to other voxels.",
    )

    fixedmodq = traits.List(
        traits.Float,
        argstr="-fixedmod %s",
        minlen=4,
        maxlen=4,
        desc="Specifies <M> <N> <Q> <tau> a spherical acquisition scheme with M measurements "
        "with q=0 and N measurements with :math:`|q|=Q` and diffusion time tau. "
        "The N measurements with :math:`|q|=Q` have unique directions. The program reads in "
        "the directions from the files in directory PointSets.",
    )

    fixedbvalue = traits.List(
        traits.Float,
        argstr="-fixedbvalue %s",
        minlen=3,
        maxlen=3,
        desc="As above, but specifies <M> <N> <b>. The resulting scheme is the same whether "
        "you specify b directly or indirectly using -fixedmodq.",
    )

    tau = traits.Float(
        argstr="-tau %G",
        desc="Sets the diffusion time separately. This overrides the diffusion time "
        "specified in a scheme file or by a scheme index for both the acquisition scheme "
        "and in the data synthesis.",
    )


class ModelFitOutputSpec(TraitedSpec):
    fitted_data = File(exists=True, desc="output file of 4D volume in voxel order")


class ModelFit(StdOutCommandLine):
    """
    Fits models of the spin-displacement density to diffusion MRI measurements.

    This is an interface to various model fitting routines for diffusion MRI data that
    fit models of the spin-displacement density function. In particular, it will fit the
    diffusion tensor to a set of measurements as well as various other models including
    two or three-tensor models. The program can read input data from a file or can
    generate synthetic data using various test functions for testing and simulations.

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> fit = cmon.ModelFit()
    >>> fit.model = 'dt'
    >>> fit.inputs.scheme_file = 'A.scheme'
    >>> fit.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> fit.run()                  # doctest: +SKIP

    """

    _cmd = "modelfit"
    input_spec = ModelFitInputSpec
    output_spec = ModelFitOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["fitted_data"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_fit.Bdouble"


class DTLUTGenInputSpec(StdOutCommandLineInputSpec):
    lrange = traits.List(
        traits.Float,
        desc="Index to one-tensor LUTs. This is the ratio L1/L3 and L2 / L3."
        "The LUT is square, with half the values calculated (because L2 / L3 cannot be "
        "less than L1 / L3 by definition)."
        "The minimum must be >= 1. For comparison, a ratio L1 / L3 = 10 with L2 / L3 = 1 "
        "corresponds to an FA of 0.891, and L1 / L3 = 15 with L2 / L3 = 1 corresponds "
        "to an FA of 0.929. The default range is 1 to 10.",
        argstr="-lrange %s",
        minlen=2,
        maxlen=2,
        position=1,
        units="NA",
    )

    frange = traits.List(
        traits.Float,
        desc="Index to two-tensor LUTs. This is the fractional anisotropy"
        " of the two tensors. The default is 0.3 to 0.94",
        argstr="-frange %s",
        minlen=2,
        maxlen=2,
        position=1,
        units="NA",
    )

    step = traits.Float(
        argstr="-step %f",
        units="NA",
        desc="Distance between points in the LUT."
        "For example, if lrange is 1 to 10 and the step is 0.1, LUT entries will be computed "
        "at L1 / L3 = 1, 1.1, 1.2 ... 10.0 and at L2 / L3 = 1.0, 1.1 ... L1 / L3."
        "For single tensor LUTs, the default step is 0.2, for two-tensor LUTs it is 0.02.",
    )

    samples = traits.Int(
        argstr="-samples %d",
        units="NA",
        desc="The number of synthetic measurements to generate at each point in the LUT. "
        "The default is 2000.",
    )

    snr = traits.Float(
        argstr="-snr %f",
        units="NA",
        desc="The signal to noise ratio of the unweighted (q = 0) measurements."
        "This should match the SNR (in white matter) of the images that the LUTs are used with.",
    )

    bingham = traits.Bool(
        argstr="-bingham",
        desc="Compute a LUT for the Bingham PDF. This is the default.",
    )

    acg = traits.Bool(argstr="-acg", desc="Compute a LUT for the ACG PDF.")

    watson = traits.Bool(argstr="-watson", desc="Compute a LUT for the Watson PDF.")

    inversion = traits.Int(
        argstr="-inversion %d",
        units="NA",
        desc="Index of the inversion to use. The default is 1 (linear single tensor inversion).",
    )

    trace = traits.Float(
        argstr="-trace %G",
        units="NA",
        desc="Trace of the diffusion tensor(s) used in the test function in the LUT generation. "
        "The default is 2100E-12 m^2 s^-1.",
    )

    scheme_file = File(
        argstr="-schemefile %s",
        mandatory=True,
        position=2,
        desc="The scheme file of the images to be processed using this LUT.",
    )


class DTLUTGenOutputSpec(TraitedSpec):
    dtLUT = File(exists=True, desc="Lookup Table")


class DTLUTGen(StdOutCommandLine):
    """
    Calibrates the PDFs for PICo probabilistic tractography.

    This program needs to be run once for every acquisition scheme.
    It outputs a lookup table that is used by the dtpicoparams program to find PICo PDF
    parameters for an image.
    The default single tensor LUT contains parameters of the Bingham distribution and is
    generated by supplying a scheme file and an estimated signal to noise in white matter
    regions of the (q=0) image.
    The default inversion is linear (inversion index 1).

    Advanced users can control several options, including the extent and resolution of the LUT,
    the inversion index, and the type of PDF. See dtlutgen(1) for details.

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> dtl = cmon.DTLUTGen()
    >>> dtl.inputs.snr = 16
    >>> dtl.inputs.scheme_file = 'A.scheme'
    >>> dtl.run()                  # doctest: +SKIP

    """

    _cmd = "dtlutgen"
    input_spec = DTLUTGenInputSpec
    output_spec = DTLUTGenOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["dtLUT"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.scheme_file)
        return name + ".dat"


class PicoPDFsInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="< %s",
        mandatory=True,
        position=1,
        desc="voxel-order data filename",
    )

    inputmodel = traits.Enum(
        "dt",
        "multitensor",
        "pds",
        argstr="-inputmodel %s",
        position=2,
        desc="input model type",
        usedefault=True,
    )

    luts = InputMultiPath(
        File(exists=True),
        argstr="-luts %s",
        mandatory=True,
        desc="Files containing the lookup tables."
        "For tensor data, one lut must be specified for each type of inversion used in the "
        "image (one-tensor, two-tensor, three-tensor)."
        "For pds, the number of LUTs must match -numpds (it is acceptable to use the same "
        "LUT several times - see example, above)."
        "These LUTs may be generated with dtlutgen.",
    )

    pdf = traits.Enum(
        "bingham",
        "watson",
        "acg",
        argstr="-pdf %s",
        position=4,
        desc="""\
Specifies the PDF to use. There are three choices:

  * watson - The Watson distribution. This distribution is rotationally symmetric.
  * bingham - The Bingham distributionn, which allows elliptical probability density contours.
  * acg - The Angular Central Gaussian distribution, which also allows elliptical probability
    density contours.

""",
        usedefault=True,
    )

    directmap = traits.Bool(
        argstr="-directmap",
        desc="Only applicable when using pds as the inputmodel. Use direct mapping between "
        "the eigenvalues and the distribution parameters instead of the log of the eigenvalues.",
    )

    maxcomponents = traits.Int(
        argstr="-maxcomponents %d",
        units="NA",
        desc="The maximum number of tensor components in a voxel (default 2) for multitensor data."
        "Currently, only the default is supported, but future releases may allow the input "
        "of three-tensor data using this option.",
    )

    numpds = traits.Int(
        argstr="-numpds %d",
        units="NA",
        desc="The maximum number of PDs in a voxel (default 3) for PD data."
        "This option determines the size of the input and output voxels."
        "This means that the data file may be large enough to accomodate three or more PDs,"
        "but does not mean that any of the voxels are classified as containing three or more PDs.",
    )


class PicoPDFsOutputSpec(TraitedSpec):
    pdfs = File(exists=True, desc="path/name of 4D volume in voxel order")


class PicoPDFs(StdOutCommandLine):
    """
    Constructs a spherical PDF in each voxel for probabilistic tractography.

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> pdf = cmon.PicoPDFs()
    >>> pdf.inputs.inputmodel = 'dt'
    >>> pdf.inputs.luts = ['lut_file']
    >>> pdf.inputs.in_file = 'voxel-order_data.Bfloat'
    >>> pdf.run()                  # doctest: +SKIP

    """

    _cmd = "picopdfs"
    input_spec = PicoPDFsInputSpec
    output_spec = PicoPDFsOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["pdfs"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_pdfs.Bdouble"


class TrackInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True, argstr="-inputfile %s", position=1, desc="input data file"
    )

    seed_file = File(exists=True, argstr="-seedfile %s", position=2, desc="seed file")

    inputmodel = traits.Enum(
        "dt",
        "multitensor",
        "sfpeak",
        "pico",
        "repbs_dt",
        "repbs_multitensor",
        "ballstick",
        "wildbs_dt",
        "bayesdirac",
        "bayesdirac_dt",
        "bedpostx_dyad",
        "bedpostx",
        argstr="-inputmodel %s",
        desc="input model type",
        usedefault=True,
    )

    tracker = traits.Enum(
        "fact",
        "euler",
        "rk4",
        argstr="-tracker %s",
        desc=(
            "The tracking algorithm controls streamlines are "
            "generated from the data. The choices are: "
            "- FACT, which follows the local fibre orientation "
            "in each voxel. No interpolation is used."
            "- EULER, which uses a fixed step size along the "
            "local fibre orientation. With nearest-neighbour "
            "interpolation, this method may be very similar to "
            "FACT, except that the step size is fixed, whereas "
            "FACT steps extend to the boundary of the next voxel "
            "(distance variable depending on the entry and exit "
            "points to the voxel)."
            "- RK4: Fourth-order Runge-Kutta method. The step "
            "size is fixed, however the eventual direction of "
            "the step is determined by taking and averaging a "
            "series of partial steps."
        ),
        usedefault=True,
    )

    interpolator = traits.Enum(
        "nn",
        "prob_nn",
        "linear",
        argstr="-interpolator %s",
        desc=(
            "The interpolation algorithm determines how "
            "the fiber orientation(s) are defined at a given "
            "continuous point within the input image. "
            "Interpolators are only used when the tracking "
            "algorithm is not FACT. The choices are: "
            "- NN: Nearest-neighbour interpolation, just "
            "uses the local voxel data directly."
            "- PROB_NN: Probabilistic nearest-neighbor "
            "interpolation,  similar  to the method pro- "
            "posed by Behrens et al [Magnetic Resonance "
            "in Medicine, 50:1077-1088, 2003]. The data "
            "is not interpolated, but at each point we "
            "randomly choose one of the 8 voxels sur- "
            "rounding a point. The probability of choosing "
            "a particular voxel is based on how close the "
            "point is to the centre of that voxel."
            "- LINEAR: Linear interpolation of the vector "
            "field containing the principal directions at "
            "each point."
        ),
    )

    stepsize = traits.Float(
        argstr="-stepsize %f",
        requires=["tracker"],
        desc=("Step size for EULER and RK4 tracking. " "The default is 1mm."),
    )

    inputdatatype = traits.Enum(
        "float", "double", argstr="-inputdatatype %s", desc="input file type"
    )

    gzip = traits.Bool(argstr="-gzip", desc="save the output image in gzip format")

    maxcomponents = traits.Int(
        argstr="-maxcomponents %d",
        units="NA",
        desc=(
            "The maximum number of tensor components in a "
            "voxel. This determines the size of the input "
            "file and does not say anything about the "
            "voxel classification. The default is 2 if "
            "the input model is multitensor and 1 if the "
            "input model is dt."
        ),
    )

    numpds = traits.Int(
        argstr="-numpds %d",
        units="NA",
        desc=(
            "The maximum number of PDs in a voxel for input "
            "models sfpeak and pico. The default is 3 for input "
            "model sfpeak and 1 for input model pico. This option "
            "determines the size of the voxels in the input file "
            "and does not affect tracking. For tensor data, use "
            "the -maxcomponents option."
        ),
    )

    data_dims = traits.List(
        traits.Int,
        desc="data dimensions in voxels",
        argstr="-datadims %s",
        minlen=3,
        maxlen=3,
        units="voxels",
    )

    voxel_dims = traits.List(
        traits.Float,
        desc="voxel dimensions in mm",
        argstr="-voxeldims %s",
        minlen=3,
        maxlen=3,
        units="mm",
    )

    ipthresh = traits.Float(
        argstr="-ipthresh %f",
        desc=(
            "Curvature threshold for tracking, expressed as "
            "the minimum dot product between two streamline "
            "orientations calculated over the length of a "
            "voxel. If the dot product between the previous "
            "and current directions is less than this "
            "threshold, then the streamline terminates. The "
            "default setting will terminate fibres that curve "
            "by more than 80 degrees. Set this to -1.0 to "
            "disable curvature checking completely."
        ),
    )

    curvethresh = traits.Float(
        argstr="-curvethresh %f",
        desc=(
            "Curvature threshold for tracking, expressed "
            "as the maximum angle (in degrees) between "
            "between two streamline orientations "
            "calculated over the length of a voxel. If "
            "the angle is greater than this, then the "
            "streamline terminates."
        ),
    )

    curveinterval = traits.Float(
        argstr="-curveinterval %f",
        requires=["curvethresh"],
        desc=(
            "Interval over which the curvature threshold "
            "should be evaluated, in mm. The default is "
            "5mm. When using the default curvature "
            "threshold of 90 degrees, this means that "
            "streamlines will terminate if they curve by "
            "more than  90  degrees over a path length "
            "of 5mm."
        ),
    )

    anisthresh = traits.Float(
        argstr="-anisthresh %f",
        desc=(
            "Terminate fibres that enter a voxel with lower "
            "anisotropy than the threshold."
        ),
    )

    anisfile = File(
        argstr="-anisfile %s",
        exists=True,
        desc=(
            "File containing the anisotropy map. This is required to "
            "apply an anisotropy threshold with non tensor data. If "
            "the map issupplied it is always used, even in tensor "
            "data."
        ),
    )

    outputtracts = traits.Enum(
        "float",
        "double",
        "oogl",
        argstr="-outputtracts %s",
        desc="output tract file type",
    )

    out_file = File(
        argstr="-outputfile %s", position=-1, genfile=True, desc="output data file"
    )

    output_root = File(
        exists=False,
        argstr="-outputroot %s",
        position=-1,
        desc="root directory for output",
    )


class TrackOutputSpec(TraitedSpec):
    tracked = File(exists=True, desc="output file containing reconstructed tracts")


class Track(CommandLine):
    """
    Performs tractography using one of the following models:
    dt', 'multitensor', 'pds', 'pico', 'bootstrap', 'ballstick', 'bayesdirac'

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> track = cmon.Track()
    >>> track.inputs.inputmodel = 'dt'
    >>> track.inputs.in_file = 'data.Bfloat'
    >>> track.inputs.seed_file = 'seed_mask.nii'
    >>> track.run()                  # doctest: +SKIP

    """

    _cmd = "track"

    input_spec = TrackInputSpec
    output_spec = TrackOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        if isdefined(self.inputs.out_file):
            out_file_path = os.path.abspath(self.inputs.out_file)
        else:
            out_file_path = os.path.abspath(self._gen_outfilename())
        outputs["tracked"] = out_file_path
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        # Currently in_file is only undefined for bedpostx input
        if not isdefined(self.inputs.in_file):
            name = "bedpostx"
        else:
            _, name, _ = split_filename(self.inputs.in_file)
        return name + "_tracked"


class TrackDT(Track):
    """
    Performs streamline tractography using tensor data

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> track = cmon.TrackDT()
    >>> track.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> track.inputs.seed_file = 'seed_mask.nii'
    >>> track.run()                 # doctest: +SKIP

    """

    def __init__(self, command=None, **inputs):
        inputs["inputmodel"] = "dt"
        return super(TrackDT, self).__init__(command, **inputs)


class TrackPICoInputSpec(TrackInputSpec):
    pdf = traits.Enum(
        "bingham",
        "watson",
        "acg",
        argstr="-pdf %s",
        desc='Specifies the model for PICo parameters. The default is "bingham.',
    )

    iterations = traits.Int(
        argstr="-iterations %d",
        units="NA",
        desc="Number of streamlines to generate at each seed point. The default is 5000.",
    )


class TrackPICo(Track):
    """
    Performs streamline tractography using Probabilistic Index of Connectivity (PICo).

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> track = cmon.TrackPICo()
    >>> track.inputs.in_file = 'pdfs.Bfloat'
    >>> track.inputs.seed_file = 'seed_mask.nii'
    >>> track.run()                  # doctest: +SKIP

    """

    input_spec = TrackPICoInputSpec

    def __init__(self, command=None, **inputs):
        inputs["inputmodel"] = "pico"
        return super(TrackPICo, self).__init__(command, **inputs)


class TrackBedpostxDeterInputSpec(TrackInputSpec):
    bedpostxdir = Directory(
        argstr="-bedpostxdir %s",
        mandatory=True,
        exists=True,
        desc=("Directory containing bedpostx output"),
    )

    min_vol_frac = traits.Float(
        argstr="-bedpostxminf %d",
        units="NA",
        desc=(
            "Zeros out compartments in bedpostx data "
            "with a mean volume fraction f of less than "
            "min_vol_frac.  The default is 0.01."
        ),
    )


class TrackBedpostxDeter(Track):
    """
    Data from FSL's bedpostx can be imported into Camino for deterministic tracking.
    (Use TrackBedpostxProba for bedpostx probabilistic tractography.)

    The tracking is based on the vector images dyads1.nii.gz, ... , dyadsN.nii.gz,
    where there are a maximum of N compartments (corresponding to each fiber
    population) in each voxel.

    It also uses the N images mean_f1samples.nii.gz, ..., mean_fNsamples.nii.gz,
    normalized such that the sum of all compartments is 1. Compartments where the
    mean_f is less than a threshold are discarded and not used for tracking.
    The default value is 0.01. This can be changed with the min_vol_frac option.

    Example
    -------
    >>> import nipype.interfaces.camino as cam
    >>> track = cam.TrackBedpostxDeter()
    >>> track.inputs.bedpostxdir = 'bedpostxout'
    >>> track.inputs.seed_file = 'seed_mask.nii'
    >>> track.run()                  # doctest: +SKIP

    """

    input_spec = TrackBedpostxDeterInputSpec

    def __init__(self, command=None, **inputs):
        inputs["inputmodel"] = "bedpostx_dyad"
        return super(TrackBedpostxDeter, self).__init__(command, **inputs)


class TrackBedpostxProbaInputSpec(TrackInputSpec):
    bedpostxdir = Directory(
        argstr="-bedpostxdir %s",
        mandatory=True,
        exists=True,
        desc=("Directory containing bedpostx output"),
    )

    min_vol_frac = traits.Float(
        argstr="-bedpostxminf %d",
        units="NA",
        desc=(
            "Zeros out compartments in bedpostx data "
            "with a mean volume fraction f of less than "
            "min_vol_frac.  The default is 0.01."
        ),
    )

    iterations = traits.Int(
        argstr="-iterations %d",
        units="NA",
        desc="Number of streamlines to generate at each seed point. The default is 1.",
    )


class TrackBedpostxProba(Track):
    """
    Data from FSL's bedpostx can be imported into Camino for probabilistic tracking.
    (Use TrackBedpostxDeter for bedpostx deterministic tractography.)

    The tracking uses the files merged_th1samples.nii.gz, merged_ph1samples.nii.gz,
    ... , merged_thNsamples.nii.gz, merged_phNsamples.nii.gz where there are a
    maximum of N compartments (corresponding to each fiber population) in each
    voxel. These images contain M samples of theta and phi, the polar coordinates
    describing the "stick" for each compartment. At each iteration, a random number
    X between 1 and M is drawn and the Xth samples of theta and phi become the
    principal directions in the voxel.

    It also uses the N images mean_f1samples.nii.gz, ..., mean_fNsamples.nii.gz,
    normalized such that the sum of all compartments is 1. Compartments where the
    mean_f is less than a threshold are discarded and not used for tracking.
    The default value is 0.01. This can be changed with the min_vol_frac option.

    Example
    -------
    >>> import nipype.interfaces.camino as cam
    >>> track = cam.TrackBedpostxProba()
    >>> track.inputs.bedpostxdir = 'bedpostxout'
    >>> track.inputs.seed_file = 'seed_mask.nii'
    >>> track.inputs.iterations = 100
    >>> track.run()                  # doctest: +SKIP

    """

    input_spec = TrackBedpostxProbaInputSpec

    def __init__(self, command=None, **inputs):
        inputs["inputmodel"] = "bedpostx"
        return super(TrackBedpostxProba, self).__init__(command, **inputs)


class TrackBayesDiracInputSpec(TrackInputSpec):
    scheme_file = File(
        argstr="-schemefile %s",
        mandatory=True,
        exists=True,
        desc=("The scheme file corresponding to the data being " "processed."),
    )

    iterations = traits.Int(
        argstr="-iterations %d",
        units="NA",
        desc=(
            "Number of streamlines to generate at each "
            "seed point. The default is 5000."
        ),
    )

    pdf = traits.Enum(
        "bingham",
        "watson",
        "acg",
        argstr="-pdf %s",
        desc="Specifies the model for PICo priors (not the curvature priors). "
        "The default is 'bingham'.",
    )

    pointset = traits.Int(
        argstr="-pointset %s",
        desc="""\
Index to the point set to use for Bayesian likelihood calculation. The index
specifies a set of evenly distributed points on the unit sphere, where each point x
defines two possible step directions (x or -x) for the streamline path. A larger
number indexes a larger point set, which gives higher angular resolution at the
expense of computation time. The default is index 1, which gives 1922 points, index 0
gives 1082 points, index 2 gives 3002 points.""",
    )

    datamodel = traits.Enum(
        "cylsymmdt",
        "ballstick",
        argstr="-datamodel %s",
        desc="""\
Model of the data for Bayesian tracking. The default model is "cylsymmdt", a diffusion
tensor with cylindrical symmetry about e_1, ie L1 >= L_2 = L_3. The other model is
"ballstick", the partial volume model (see ballstickfit).""",
    )

    curvepriork = traits.Float(
        argstr="-curvepriork %G",
        desc="""\
Concentration parameter for the prior distribution on fibre orientations given the fibre
orientation at the previous step. Larger values of k make curvature less likely.""",
    )

    curvepriorg = traits.Float(
        argstr="-curvepriorg %G",
        desc="""\
Concentration parameter for the prior distribution on fibre orientations given
the fibre orientation at the previous step. Larger values of g make curvature less likely.""",
    )

    extpriorfile = File(
        exists=True,
        argstr="-extpriorfile %s",
        desc="""\
Path to a PICo image produced by picopdfs. The PDF in each voxel is used as a prior for
the fibre orientation in Bayesian tracking. The prior image must be in the same space
as the diffusion data.""",
    )

    extpriordatatype = traits.Enum(
        "float",
        "double",
        argstr="-extpriordatatype %s",
        desc='Datatype of the prior image. The default is "double".',
    )


class TrackBayesDirac(Track):
    """
    Perform streamline tractography using a Bayesian tracking with Dirac priors.

    Example
    -------

    >>> import nipype.interfaces.camino as cmon
    >>> track = cmon.TrackBayesDirac()
    >>> track.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> track.inputs.seed_file = 'seed_mask.nii'
    >>> track.inputs.scheme_file = 'bvecs.scheme'
    >>> track.run()                  # doctest: +SKIP

    """

    input_spec = TrackBayesDiracInputSpec

    def __init__(self, command=None, **inputs):
        inputs["inputmodel"] = "bayesdirac"
        return super(TrackBayesDirac, self).__init__(command, **inputs)


class TrackBallStick(Track):
    """
    Performs streamline tractography using ball-stick fitted data

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> track = cmon.TrackBallStick()
    >>> track.inputs.in_file = 'ballstickfit_data.Bfloat'
    >>> track.inputs.seed_file = 'seed_mask.nii'
    >>> track.run()                  # doctest: +SKIP

    """

    def __init__(self, command=None, **inputs):
        inputs["inputmodel"] = "ballstick"
        return super(TrackBallStick, self).__init__(command, **inputs)


class TrackBootstrapInputSpec(TrackInputSpec):
    scheme_file = File(
        argstr="-schemefile %s",
        mandatory=True,
        exists=True,
        desc="The scheme file corresponding to the data being processed.",
    )

    iterations = traits.Int(
        argstr="-iterations %d",
        units="NA",
        desc="Number of streamlines to generate at each seed point.",
    )

    inversion = traits.Int(
        argstr="-inversion %s",
        desc="""\
Tensor reconstruction algorithm for repetition bootstrapping.
Default is 1 (linear reconstruction, single tensor).""",
    )

    bsdatafiles = traits.List(
        File(exists=True),
        mandatory=True,
        argstr="-bsdatafile %s",
        desc="""\
Specifies files containing raw data for repetition bootstrapping.
Use -inputfile for wild bootstrap data.""",
    )

    bgmask = File(
        argstr="-bgmask %s",
        exists=True,
        desc="""\
Provides the name of a file containing a background mask computed using, for example,
FSL's bet2 program.
The mask file contains zero in background voxels and non-zero in foreground.""",
    )


class TrackBootstrap(Track):
    """
    Performs bootstrap streamline tractography using mulitple scans of the same subject

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> track = cmon.TrackBootstrap()
    >>> track.inputs.inputmodel='repbs_dt'
    >>> track.inputs.scheme_file = 'bvecs.scheme'
    >>> track.inputs.bsdatafiles = ['fitted_data1.Bfloat', 'fitted_data2.Bfloat']
    >>> track.inputs.seed_file = 'seed_mask.nii'
    >>> track.run()                  # doctest: +SKIP

    """

    input_spec = TrackBootstrapInputSpec

    def __init__(self, command=None, **inputs):
        return super(TrackBootstrap, self).__init__(command, **inputs)


class ComputeMeanDiffusivityInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="< %s",
        mandatory=True,
        position=1,
        desc="Tensor-fitted data filename",
    )

    scheme_file = File(
        exists=True,
        argstr="%s",
        position=2,
        desc="Camino scheme file (b values / vectors, see camino.fsl2scheme)",
    )

    out_file = File(argstr="> %s", position=-1, genfile=True)

    inputmodel = traits.Enum(
        "dt",
        "twotensor",
        "threetensor",
        argstr="-inputmodel %s",
        desc="""\
Specifies the model that the input tensor data contains parameters for.
By default, the program assumes that the input data
contains a single diffusion tensor in each voxel.""",
    )

    inputdatatype = traits.Enum(
        "char",
        "short",
        "int",
        "long",
        "float",
        "double",
        argstr="-inputdatatype %s",
        desc="Specifies the data type of the input file.",
    )

    outputdatatype = traits.Enum(
        "char",
        "short",
        "int",
        "long",
        "float",
        "double",
        argstr="-outputdatatype %s",
        desc="Specifies the data type of the output data.",
    )


class ComputeMeanDiffusivityOutputSpec(TraitedSpec):
    md = File(exists=True, desc="Mean Diffusivity Map")


class ComputeMeanDiffusivity(StdOutCommandLine):
    """
    Computes the mean diffusivity (trace/3) from diffusion tensors.

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> md = cmon.ComputeMeanDiffusivity()
    >>> md.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> md.inputs.scheme_file = 'A.scheme'
    >>> md.run()                  # doctest: +SKIP

    """

    _cmd = "md"
    input_spec = ComputeMeanDiffusivityInputSpec
    output_spec = ComputeMeanDiffusivityOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["md"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_MD.img"  # Need to change to self.inputs.outputdatatype


class ComputeFractionalAnisotropyInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="< %s",
        mandatory=True,
        position=1,
        desc="Tensor-fitted data filename",
    )

    scheme_file = File(
        exists=True,
        argstr="%s",
        position=2,
        desc="Camino scheme file (b values / vectors, see camino.fsl2scheme)",
    )

    inputmodel = traits.Enum(
        "dt",
        "twotensor",
        "threetensor",
        "multitensor",
        argstr="-inputmodel %s",
        desc="""\
Specifies the model that the input tensor data contains parameters for.
By default, the program assumes that the input data
contains a single diffusion tensor in each voxel.""",
    )

    inputdatatype = traits.Enum(
        "char",
        "short",
        "int",
        "long",
        "float",
        "double",
        argstr="-inputdatatype %s",
        desc="Specifies the data type of the input file.",
    )

    outputdatatype = traits.Enum(
        "char",
        "short",
        "int",
        "long",
        "float",
        "double",
        argstr="-outputdatatype %s",
        desc="Specifies the data type of the output data.",
    )


class ComputeFractionalAnisotropyOutputSpec(TraitedSpec):
    fa = File(exists=True, desc="Fractional Anisotropy Map")


class ComputeFractionalAnisotropy(StdOutCommandLine):
    """
    Computes the fractional anisotropy of tensors.

    Reads diffusion tensor (single, two-tensor or three-tensor) data from the standard input,
    computes the fractional anisotropy (FA) of each tensor and outputs the results to the
    standard output. For multiple-tensor data the program outputs the FA of each tensor,
    so for three-tensor data, for example, the output contains three fractional anisotropy
    values per voxel.

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> fa = cmon.ComputeFractionalAnisotropy()
    >>> fa.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> fa.inputs.scheme_file = 'A.scheme'
    >>> fa.run()                  # doctest: +SKIP

    """

    _cmd = "fa"
    input_spec = ComputeFractionalAnisotropyInputSpec
    output_spec = ComputeFractionalAnisotropyOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["fa"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_FA.Bdouble"  # Need to change to self.inputs.outputdatatype


class ComputeTensorTraceInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="< %s",
        mandatory=True,
        position=1,
        desc="Tensor-fitted data filename",
    )

    scheme_file = File(
        exists=True,
        argstr="%s",
        position=2,
        desc="Camino scheme file (b values / vectors, see camino.fsl2scheme)",
    )

    inputmodel = traits.Enum(
        "dt",
        "twotensor",
        "threetensor",
        "multitensor",
        argstr="-inputmodel %s",
        desc="""\
Specifies the model that the input tensor data contains parameters for.
By default, the program assumes that the input data
contains a single diffusion tensor in each voxel.""",
    )

    inputdatatype = traits.Enum(
        "char",
        "short",
        "int",
        "long",
        "float",
        "double",
        argstr="-inputdatatype %s",
        desc="Specifies the data type of the input file.",
    )

    outputdatatype = traits.Enum(
        "char",
        "short",
        "int",
        "long",
        "float",
        "double",
        argstr="-outputdatatype %s",
        desc="Specifies the data type of the output data.",
    )


class ComputeTensorTraceOutputSpec(TraitedSpec):
    trace = File(exists=True, desc="Trace of the diffusion tensor")


class ComputeTensorTrace(StdOutCommandLine):
    """
    Computes the trace of tensors.

    Reads diffusion tensor (single, two-tensor or three-tensor) data from the standard input,
    computes the trace of each tensor, i.e., three times the mean diffusivity, and outputs
    the results to the standard output. For multiple-tensor data the program outputs the
    trace of each tensor, so for three-tensor data, for example, the output contains three
    values per voxel.

    Divide the output by three to get the mean diffusivity.

    Example
    -------
    >>> import nipype.interfaces.camino as cmon
    >>> trace = cmon.ComputeTensorTrace()
    >>> trace.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> trace.inputs.scheme_file = 'A.scheme'
    >>> trace.run()                 # doctest: +SKIP

    """

    _cmd = "trd"
    input_spec = ComputeTensorTraceInputSpec
    output_spec = ComputeTensorTraceOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["trace"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_TrD.img"  # Need to change to self.inputs.outputdatatype


class ComputeEigensystemInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="< %s",
        mandatory=True,
        position=1,
        desc="Tensor-fitted data filename",
    )

    inputmodel = traits.Enum(
        "dt",
        "multitensor",
        argstr="-inputmodel %s",
        desc="Specifies the model that the input data contains parameters for",
    )

    maxcomponents = traits.Int(
        argstr="-maxcomponents %d",
        desc="The maximum number of tensor components in a voxel of the input data.",
    )

    inputdatatype = traits.Enum(
        "double",
        "float",
        "long",
        "int",
        "short",
        "char",
        argstr="-inputdatatype %s",
        usedefault=True,
        desc=(
            "Specifies the data type of the input data. "
            "The data type can be any of the following strings: "
            '"char", "short", "int", "long", "float" or "double".'
            "Default is double data type"
        ),
    )

    outputdatatype = traits.Enum(
        "double",
        "float",
        "long",
        "int",
        "short",
        "char",
        argstr="-outputdatatype %s",
        usedefault=True,
        desc="Specifies the data type of the output data.",
    )


class ComputeEigensystemOutputSpec(TraitedSpec):
    eigen = File(exists=True, desc="Trace of the diffusion tensor")


class ComputeEigensystem(StdOutCommandLine):
    """
    Computes the eigensystem from tensor fitted data.

    Reads diffusion tensor (single, two-tensor, three-tensor or multitensor) data from the
    standard input, computes the eigenvalues and eigenvectors of each tensor and outputs the
    results to the standard output. For multiple-tensor data the program outputs the
    eigensystem of each tensor. For each tensor the program outputs: {l_1, e_11, e_12, e_13,
    l_2, e_21, e_22, e_33, l_3, e_31, e_32, e_33}, where l_1 >= l_2 >= l_3 and e_i = (e_i1,
    e_i2, e_i3) is the eigenvector with eigenvalue l_i. For three-tensor data, for example,
    the output contains thirty-six values per voxel.

    Example
    -------

    >>> import nipype.interfaces.camino as cmon
    >>> dteig = cmon.ComputeEigensystem()
    >>> dteig.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> dteig.run()                  # doctest: +SKIP
    """

    _cmd = "dteig"
    input_spec = ComputeEigensystemInputSpec
    output_spec = ComputeEigensystemOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["eigen"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        datatype = self.inputs.outputdatatype
        return name + "_eig.B" + datatype
